# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np
import itertools
import copy
import contextlib
import sys
import declarative

from .data_filtfit_base import DataFiltFitBase
from .fake_aid import ensure_aid


def norm_sq(v):
    return v.real.dot(v.real) + v.imag.dot(v.imag)


def abs_sq(x):
    return x.real**2 + x.imag**2


class MultiReprFilter(DataFiltFitBase):
    def __build__(
        self,
        _args                  = None,
        parent                 = None,
        gain_delay_coding      = None,
        num_codings            = None,
        den_codings            = None,
        parameters             = None,
        extra_poles            = None,
        extra_zeros            = None,
        residuals_type         = None,
        codings_ignore         = None,
        residuals_log_im_scale = None,
        **kwargs
    ):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")
        super(MultiReprFilter, self).__build__(
            parent = parent,
            **kwargs
        )
        if parent is not None and isinstance(parent, MultiReprFilter):
            if parameters is None:
                parameters        = parent.parameters
            if gain_delay_coding is None:
                gain_delay_coding = parent.gain_delay_coding
            #TODO, use the copier
            if num_codings is None:
                num_codings       = parent.num_codings
            #TODO, use the copier
            if den_codings is None:
                den_codings       = parent.den_codings
            if extra_poles is None:
                extra_poles       = parent.extra_poles
            if extra_zeros is None:
                extra_zeros       = parent.extra_zeros
            if residuals_type is None:
                residuals_type    = parent.residuals_type
            if codings_ignore is None:
                arg_codings_ignore = parent.codings_ignore
            if residuals_log_im_scale is None:
                residuals_log_im_scale = parent.residuals_log_im_scale
        else:
            if gain_delay_coding is None:
                raise RuntimeError("Must specify gain_delay_coding")
            if num_codings is None:
                raise RuntimeError("Must specify num_codings")
            if den_codings is None:
                raise RuntimeError("Must specify den_codings")
            if extra_poles is None:
                extra_poles = ()
            if extra_zeros is None:
                extra_zeros = ()
            if residuals_type is None:
                residuals_type = 'log'
            if codings_ignore is None:
                arg_codings_ignore = set()
            if residuals_log_im_scale is None:
                residuals_log_im_scale = 1
        arg_gain_delay_coding = gain_delay_coding
        arg_den_codings       = den_codings
        arg_num_codings       = num_codings

        #the parameters are mutable objects that need copy rules
        @self.deco_copier
        def gain_delay_coding(self, val):
            return copy.deepcopy(val)

        @self.deco_setter
        def gain_delay_coding(self, val):
            return copy.deepcopy(val)

        @self.deco_copier
        def num_codings(self, val):
            return copy.deepcopy(val)

        @self.deco_setter
        def num_codings(self, val):
            return list(copy.deepcopy(val))

        @self.deco_setter
        def den_codings(self, val):
            return list(copy.deepcopy(val))

        @self.deco_copier
        def den_codings(self, val):
            return copy.deepcopy(val)

        @self.deco_generator
        def root_codings_set(self):
            return set(self.num_codings + self.den_codings)

        self.gain_delay_coding = arg_gain_delay_coding
        self.num_codings       = arg_num_codings
        self.den_codings       = arg_den_codings
        self.extra_poles       = extra_poles
        self.extra_zeros       = extra_zeros
        self.residuals_type    = residuals_type
        self.residuals_log_im_scale = residuals_log_im_scale

        @self.deco_setter
        def codings_ignore(self, val):
            self.codings_revision = self.codings_revision + 1
            if val is None:
                val = set()
            else:
                coding_id_map = {
                    self.gain_delay_coding.coding_id : self.gain_delay_coding
                }
                for coding in self.num_codings:
                    coding_id_map[coding.coding_id] = coding
                for coding in self.den_codings:
                    coding_id_map[coding.coding_id] = coding

                codings_ignore_mapped = set()
                for coding in val:
                    mcoding = coding_id_map.get(coding.coding_id, None)
                    if mcoding is not None:
                        codings_ignore_mapped.add(mcoding)
                val = codings_ignore_mapped
            return val
        #must adjust this whenever something happens to the codings
        self.codings_revision  = 0
        self.auto_jacobian = False

        @self.deco_generator
        def coding_lists(self):
            num_active = []
            num_ignore = []
            den_active = []
            den_ignore = []
            if self.gain_delay_coding not in self.codings_ignore:
                num_active.append(self.gain_delay_coding)
            else:
                num_ignore.append(self.gain_delay_coding)
            for idx, coding in enumerate(self.num_codings):
                if coding not in self.codings_ignore:
                    num_active.append(coding)
                else:
                    num_ignore.append(coding)
            for idx, coding in enumerate(self.den_codings):
                if coding not in self.codings_ignore:
                    den_active.append(coding)
                else:
                    den_ignore.append(coding)
            return declarative.Bunch(
                num_active = num_active,
                num_ignore = num_ignore,
                den_active = den_active,
                den_ignore = den_ignore,
            )

        @self.deco_setter
        def xfer_fit(self, val):
            #access just for the dependency
            self.codings_revision
            return val

        @self.deco_generator
        def xfer_fit(self):
            #access just for the dependency
            #print("NEW_XFER")
            self.codings_revision
            if self.auto_jacobian:
                xfer, jac = self.transfer_jacobian_generate()
                self.residuals_jacobian = jac
                return xfer
            else:
                return self.transfer_generate()

        @self.deco_generator
        def xfer_inactive(self):
            #access just for the dependency
            self.codings_revision
            return self.transfer_generate(generate_inactive = True)

        @self.deco_setter
        def residuals_jacobian(self, val):
            #access just for the dependency
            self.codings_revision
            return val

        @self.deco_generator
        def residuals_jacobian(self):
            #access just for the dependency
            #print("NEW_JAC")
            self.codings_revision
            xfer, jac = self.transfer_jacobian_generate()
            #TODO can do this assignment, but need to set the dependencies with it
            self.xfer_fit = xfer
            return np.asarray(jac)

        @self.deco_generator
        def parameters(self):
            params = []
            N_params = 0
            for coding in self.coding_lists.num_active:
                N_params += coding.N_parameters
                params.extend(coding.reduce())
            for coding in self.coding_lists.den_active:
                N_params += coding.N_parameters
                params.extend(coding.reduce())
            #add the dependency
            self.dependencies('codings_revision')
            return np.asarray(params, dtype = float)

        @self.deco_setter
        def parameters(self, params):
            #access for the dependency, and update the value to clear other dependent values
            self.codings_revision = self.codings_revision + 1
            self.dependencies('codings_revision')
            if np.all(self.parameters == params):
                #print("SAME")
                return params
            N_current = 0
            for coding in self.coding_lists.num_active:
                num_next = coding.N_parameters
                coding.update(*params[N_current:N_current + num_next])
                N_current += num_next
            for coding in self.coding_lists.den_active:
                num_next = coding.N_parameters
                coding.update(*params[N_current:N_current + num_next])
                N_current += num_next
            assert(N_current == len(params))
            return params

        @self.deco_generator
        def coding_maps(self):
            #TODO, use coding_lists, but have to handle gain_delay_coding
            mapping = []
            locations_map = {}
            parameters_map = {}
            N_current = 0
            if self.gain_delay_coding not in self.codings_ignore:
                mapping.extend(self.gain_delay_coding.N_parameters * [self.gain_delay_coding])
                num_next = self.gain_delay_coding.N_parameters
                parameters_map[self.gain_delay_coding] = N_current
                N_current += num_next
                locations_map[self.gain_delay_coding] = 'gain_delay_coding'
            num_idx = N_current
            for idx, coding in enumerate(self.num_codings):
                if coding in self.codings_ignore:
                    continue
                num_next = coding.N_parameters
                parameters_map[coding] = N_current
                N_current += num_next
                mapping.extend(coding.N_parameters * [coding])
                locations_map[coding] = ('num_codings', idx)
            den_idx = N_current
            for idx, coding in enumerate(self.den_codings):
                if coding in self.codings_ignore:
                    continue
                num_next = coding.N_parameters
                parameters_map[coding] = N_current
                N_current += num_next
                mapping.extend(coding.N_parameters * [coding])
                locations_map[coding] = ('den_codings', idx)
            #access for the dependency
            self.codings_revision
            return declarative.Bunch(
                p2c     = mapping,
                c2loc   = locations_map,
                c2p     = parameters_map,
                num_idx = num_idx,
                den_idx = den_idx,
            )

        @self.deco_generator
        def zeros(self):
            #access just for the dependency
            self.codings_revision
            return np.asarray([r for coding in self.num_codings for r in coding.roots(self)], dtype = complex)

        @self.deco_generator
        def poles(self):
            #access just for the dependency
            self.codings_revision
            return np.asarray([r for coding in self.den_codings for r in coding.roots(self)], dtype = complex)

        @self.deco_generator
        def zeros_full(self):
            #access just for the dependency
            self.codings_revision
            return np.concatenate([self.zeros, self.extra_zeros])

        @self.deco_generator
        def poles_full(self):
            #access just for the dependency
            self.codings_revision
            return np.concatenate([self.poles, self.extra_poles])

        @self.deco_setter
        def zeros(self, val):
            raise RuntimeError("Zeros not Directly Settable")

        @self.deco_setter
        def poles(self, val):
            raise RuntimeError("Poles not Directly Settable")

        @self.deco_generator
        def residuals(self):
            retB = declarative.Bunch()
            R = self.xfer_fit/self.data
            retB.resP = self.W * (R - 1)
            retB.resZ = self.W * (1/R - 1)
            retB.resD = self.W * (R - 1/R)

            R_abs_sq = abs_sq(R)
            R_abs = R_abs_sq**.5
            log_re = self.W * np.log(R_abs)
            log_im = self.W * R.imag / R_abs
            retB.resL = log_re + 1j * log_im * self.residuals_log_im_scale
            return retB

        @self.deco_generator
        def residuals_preferred(self):
            return self.residuals_compute(self.data, self.xfer_fit, self.W)

        @self.deco_generator
        def residuals_total(self):
            return np.sum(abs_sq(self.residuals_preferred)) / (2 * len(self.data) - len(self.parameters))

        @self.deco_generator
        def residuals_max(self):
            #includes adjustment for parameter number. Not a real thing statistically...
            return np.max(abs_sq(self.residuals_preferred))

        @self.deco_generator
        def residuals_med(self):
            #includes adjustment for parameter number. Not a real thing statistically...
            return np.median(abs_sq(self.residuals_preferred))

        @self.deco_generator
        def fisher_matrix(self):
            #TODO, compute this using QR decomposition first
            #followed perhaps by SVD
            pjac = self.residuals_jacobian
            phess = np.dot(pjac, pjac.conjugate().T).real
            return phess

        @self.deco_generator
        def fisher_matrix_rel(self):
            #TODO, compute this using QR decomposition first
            #followed perhaps by SVD
            mat = np.copy(self.fisher_matrix)
            d = mat.diagonal()**.5

            #recondition to prevent NANs
            d[d < 1e-15] = 1

            mat = mat / d.reshape(-1, 1)
            mat = mat / d.reshape(1, -1)
            return mat

        self.rcond_cov = 1e-5
        @self.deco_generator
        def cov_matrix(self):
            #cov = np.linalg.pinv(self.fisher_matrix, rcond=1e-10)
            #uses the reconditioned fisher_matrix. This allows it to keep even the very badly behaved
            #values
            d = self.fisher_matrix.diagonal()**.5
            #recondition to prevent NANs
            d[d < 1e-15] = 1
            #TODO, find optimal rcond!
            p = np.linalg.pinv(
                self.fisher_matrix_rel,
                rcond = self.rcond_cov,
            )
            cov = p / d.reshape(1, -1) / d.reshape(-1, 1)
            #uses only the length of resP, so resZ would be the same
            return cov * self.residuals_total / (len(self.residuals_preferred) * 2 - cov.shape[0])

        @self.deco_generator
        def cor_matrix(self):
            cov = np.copy(self.cov_matrix)
            d = cov.diagonal()**.5
            cov = cov / d.reshape(-1, 1)
            cov = cov / d.reshape(1, -1)
            return cov

        @self.deco_generator
        def xfer_fit_error(self):
            cov = self.cov_matrix
            xfer, dfjac = self.transfer_jacobian_generate(xfer_fit_jacobian = True)
            dfjac = np.asarray(dfjac)
            perr_r = np.einsum(str("ij,ii,ij->j"), dfjac.real, cov, dfjac.real)
            perr_i = np.einsum(str("ij,ii,ij->j"), dfjac.imag, cov, dfjac.imag)

            dfjacAmp = np.einsum(str("ij,j->ij"), dfjac, abs(xfer) / xfer)
            perr_a = np.einsum(str("ij,ii,ij->j"), dfjacAmp.real, cov, dfjacAmp.real)
            dfjacRel = np.einsum(str("ij,j->ij"), dfjac, 1 / xfer)
            perr_ar = np.einsum(str("ij,ii,ij->j"), dfjacRel.real, cov, dfjacRel.real)
            perr_p = np.einsum(str("ij,ii,ij->j"), dfjacRel.imag, cov, dfjacRel.imag)
            cplx = perr_r**.5 + 1j * perr_i**.5
            return declarative.Bunch(
                cplx  = cplx,
                mag   = perr_a**.5,
                mag_rel  = perr_ar**.5,
                mag_rel2 = perr_a**.5 / abs(xfer),
                phase = perr_p**.5,
            )

        @self.deco_generator
        def residuals_error(self):
            cov = self.cov_matrix
            res = self.residuals_preferred
            dfjac = self.residuals_jacobian
            dfjac = np.asarray(dfjac)
            perr_r = np.einsum(str("ij,ii,ij->j"), dfjac.real, cov, dfjac.real)
            perr_i = np.einsum(str("ij,ii,ij->j"), dfjac.imag, cov, dfjac.imag)

            dfjacAmp = np.einsum(str("ij,j->ij"), dfjac, abs(res) / res)
            perr_a = np.einsum(str("ij,ii,ij->j"), dfjacAmp.real, cov, dfjacAmp.real)
            dfjacRel = np.einsum(str("ij,j->ij"), dfjac, 1 / res)
            perr_ar = np.einsum(str("ij,ii,ij->j"), dfjacRel.real, cov, dfjacRel.real)
            perr_p = np.einsum(str("ij,ii,ij->j"), dfjacRel.imag, cov, dfjacRel.imag)
            cplx = perr_r**.5 + 1j * perr_i**.5
            return declarative.Bunch(
                cplx  = cplx,
                mag   = perr_a**.5,
                mag_rel   = perr_ar**.5,
                phase = perr_p**.5,
            )

        #assign after the setter is built
        self.codings_ignore = arg_codings_ignore
        return  # ~__init__

    def transfer_generate(self, generate_inactive = False):
        if generate_inactive:
            #I believe these need a phasing adjustment if extra_poles and extra_zeros have different sizes
            #TODO also, if there are many extra_poles and extra_zeros, this is not numerically stable (should interlace)
            xfer = np.polynomial.polynomial.polyvalfromroots(self.Xn_grid, self.extra_zeros)
            xfer = xfer / np.polynomial.polynomial.polyvalfromroots(self.Xn_grid, self.extra_poles)
            for coding_n, coding_d in itertools.zip_longest(
                    self.coding_lists.num_ignore,
                    self.coding_lists.den_ignore,
                    fillvalue = None,
            ):
                if coding_n is not None:
                    xfer = xfer * coding_n.transfer(self)
                if coding_d is not None:
                    xfer = xfer / coding_d.transfer(self)
            return xfer
        else:
            xfer = self.xfer_inactive
            for coding_n, coding_d in itertools.zip_longest(
                    self.coding_lists.num_active,
                    self.coding_lists.den_active,
                    fillvalue = None,
            ):
                if coding_n is not None:
                    xfer = xfer * coding_n.transfer(self)
                if coding_d is not None:
                    xfer = xfer / coding_d.transfer(self)
            return xfer

    def copy(self, **kwargs):
        return self.__class__(
            parent = self,
            **kwargs
        )

    def residuals_compute(self, data, xfer, W = 1):
        rtype = self.residuals_type
        R = xfer/data
        if rtype == 'poles':
            return W * (R - 1)
        elif rtype == 'zeros':
            return W * (1/R - 1)
        elif rtype == 'dualA':
            return W * (R/2 + 0.5/R - 1)
        elif rtype == 'dualB':
            return W * (R - 1/R)
        elif rtype == 'log':
            R_abs_sq = abs_sq(R)
            R_abs = R_abs_sq**.5
            select = (R_abs <= 0) | ~np.isfinite(R_abs)
            if any(select):
                print("BADNESS AT: ", self.F_Hz[select], R_abs[select], np.sum(select) / len(select), file = sys.stderr)
                print(self.gain_delay_coding.p_gain, self.gain_delay_coding.p_delay, file = sys.stderr)
            log_re = W * np.log(R_abs)
            log_im = W * R.imag / R_abs
            return log_re + 1j * log_im * self.residuals_log_im_scale
        return

    def transfer_jacobian_generate(
        self,
        xfer_fit_jacobian = False,
    ):
        #in principle it is faster to generate the xfer with the jacobian,
        #I am separating them as a wash
        # NEEDS SAME ORDERING AS PARAMETERS LIST

        #use separate bins for num and den so that the interlacing doesn't screw up the order
        op = declarative.Bunch(
            jac_list_num = [],
            jac_list_den = [],
            xfer_running = 1
        )
        op.xfer_running = self.xfer_inactive
        rtype = self.residuals_type

        if xfer_fit_jacobian or rtype in ['poles', 'zeros', 'dualB', 'dualA']:
            def jac_modify(pjac, xfer, num):
                def generate(XY):
                    if num:
                        return XY * pjac / xfer
                    else:
                        return -XY * pjac / xfer
                return generate

        elif rtype == 'log':
            def jac_modify(pjac, xfer, num):
                jac_abs_sq = abs_sq(xfer)
                jac_re = (pjac.real * xfer.real + pjac.imag * xfer.imag) / jac_abs_sq

                def generate(XY):
                    Y = XY / xfer
                    XY_abs = abs(XY)
                    jac_im = self.residuals_log_im_scale * ((pjac.imag * Y.real + pjac.real * Y.imag) - XY.imag * jac_re) / XY_abs
                    if num:
                        return jac_re + 1j * jac_im
                    else:
                        return -(jac_re + 1j * jac_im)
                return generate

        def coding_jac_append(coding, num):
            #don't include in jacobian if the codings_ignore contains the coding
            xfer, jac = coding.derivative_wtrans(self)

            if num:
                op.xfer_running = op.xfer_running * xfer
                for pjac in jac:
                    op.jac_list_num.append(jac_modify(pjac, xfer, num = True))
            else:
                op.xfer_running = op.xfer_running / xfer
                for pjac in jac:
                    op.jac_list_den.append(jac_modify(pjac, xfer, num = False))

        #interlace to prevent the value becoming crazy small and numerically unstable
        for coding_n, coding_d in itertools.zip_longest(
                self.coding_lists.num_active,
                self.coding_lists.den_active,
                fillvalue = None,
        ):
            if coding_n is not None:
                coding_jac_append(coding_n, num = True)
            if coding_d is not None:
                coding_jac_append(coding_d, num = False)

        xfer = op.xfer_running
        #have to do this after, otherwise they will be out of order from the interlacing
        jac_list = op.jac_list_num + op.jac_list_den
        #now need to actually inject the jacobian side of the chain rule, with the proper signs for the num, denom
        #first set are numerators of the xfer, including gain_delay_coding
        if not xfer_fit_jacobian:
            if rtype == 'poles':
                XY = (xfer/self.data)
            elif rtype == 'zeros':
                #minus sign to account for xfer on the denominator
                XY = -(self.data/xfer)
            elif rtype == 'dualA':
                #minus sign goes to plus on the second term
                XY = (xfer/self.data - self.data/xfer)/2
            elif rtype == 'dualB':
                #minus sign goes to plus on the second term
                XY = (xfer/self.data + self.data/xfer)
            elif rtype == 'log':
                #minus sign goes to plus on the second term
                XY = xfer/self.data
        else:
            XY = xfer

        for idx in range(len(jac_list)):
            jac_list[idx] = self.W * jac_list[idx](XY)

        return xfer, jac_list

    def optimize(
        self,
        **kwargs
    ):
        #start with Nelder-Mead as it is fast, not needing the jacobian, then use Levenberg-Marquardt to seal the deal
        #self.optimize_NM(**kwargs)
        return self.optimize_LS(**kwargs)

    opt_xtol_default    = 3e-16
    opt_ftol_default    = 3e-5
    opt_gtol_default    = 3e-5
    opt_method_default  = 'trf'
    opt_rescale_default = True
    opt_nfev_default    = 1000

    def optimize_LS(
        self,
        xtol           = None,
        ftol           = None,
        gtol           = None,
        method         = None,
        x_scale        = 'jac',
        max_nfev       = None,
        residuals_type = None,
        rescale        = True,
        aid            = None,
        **kwargs
    ):
        aid = ensure_aid(aid)

        args = [
            'optimize_LS_{arg}.MRF',
            'optimize_{arg}.MRF',
            'optimize_LS_{arg}',
            'optimize_{arg}',
        ]
        xtol     = aid.hint_arg(xtol,     args, self.opt_xtol_default, arg = 'xtol')
        ftol     = aid.hint_arg(ftol,     args, self.opt_ftol_default, arg = 'ftol')
        gtol     = aid.hint_arg(gtol,     args, self.opt_gtol_default, arg = 'gtol')
        max_nfev = aid.hint_arg(max_nfev, args, self.opt_nfev_default, arg = 'max_nfev')

        method  = method if method is not None else self.opt_method_default
        rescale = rescale if rescale is not None else self.opt_rescale_default

        res_type_prev = self.residuals_type
        if residuals_type is not None:
            self.residuals_type = residuals_type
        if rescale:
            p_rescale = np.abs(np.asarray(self.parameters)) + 1e-5
        else:
            p_rescale = np.asarray(1)
        #TODO consider using auto_jacobian, gotta profile to see if it helps
        if abs(self.residuals_log_im_scale) > 1e-12:
            def opt(p):
                self.parameters = p * p_rescale
                return np.concatenate([
                    self.residuals_preferred.real,
                    self.residuals_preferred.imag,
                ])

            def opt_j(p):
                self.parameters = p * p_rescale
                jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
                return np.concatenate([
                    jac.real,
                    jac.imag
                ], axis = 1).T
        else:
            def opt(p):
                self.parameters = p * p_rescale
                return np.concatenate([
                    self.residuals_preferred.real,
                ])

            def opt_j(p):
                self.parameters = p * p_rescale
                jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
                return np.concatenate([
                    jac.real,
                ], axis = 1).T

        #TODO, this is here because matlab can't seem to import it. 
        import scipy.optimize
        opt_r = scipy.optimize.least_squares(
            opt,
            np.asarray(self.parameters) / p_rescale,
            jac      = opt_j,
            xtol     = xtol,
            ftol     = ftol,
            gtol     = gtol,
            x_scale  = x_scale,
            #method   = method,
            max_nfev = max_nfev,
            **kwargs
        )
        #don't actually return an OptimizeResult because it kills matlab
        results = declarative.Bunch()
        results.update(opt_r)
        self.parameters = results.x * p_rescale
        aid.log(results.message, results.nfev, results.njev)
        if residuals_type is not None:
            self.residuals_type = res_type_prev
        return results

    def optimize_NM(
        self,
        bootstrap = True,
        residuals_type = None,
        **kwargs
    ):
        """
        Uses Nelder-Mead with a simplex that is bootstrapped from the svd of the jacobian. This gives particularly good initial directions to check
        """
        res_type_prev = self.residuals_type
        if residuals_type is not None:
            self.residuals_type = residuals_type
        try:
            if bootstrap:
                jac = np.concatenate([
                    self.residuals_jacobian.real,
                    self.residuals_jacobian.imag
                ], axis = 1)
                res = np.concatenate([
                    self.residuals_preferred.real,
                    self.residuals_preferred.imag
                ], axis = 0)
                d = np.sum(jac * jac, axis = 1)**.5
                #p_rescale the jacobian using the norms, makes the sum of the singular values equal to the number
                #of parameters (and greatly improves conditioning)
                d[abs(d) < 1e-10] = 1
                jac /= d.reshape(-1, 1)
                u, s, v = np.linalg.svd(
                    jac,
                    full_matrices=False
                )
                n_r = (res.dot(res))**.5
                v_r = v.dot(res) / n_r
                #deweight the smallest singular value as it will be replaced and it shouldn't screw up the fit if it is nasty
                #TODO, handle bad singular values
                u_proj = u * (v_r * n_r / (s + 1e-4 + np.min(s)) / d).reshape(-1, 1)
                u_best = np.sum(u_proj, axis = 1)
                #now replace the smallest singular vector with the optimal nearest quadratic solution
                u_proj[:, -1] = u_best
                u_use = np.concatenate([np.zeros_like(d).reshape(-1, 1), u_proj], axis = 1).T + np.asarray(self.parameters)
                return self.optimize_min(
                    initial_simplex = u_use,
                    **kwargs
                )
            else:
                return self.optimize_min(
                    initial_simplex = u_use,
                    **kwargs
                )
        finally:
            if residuals_type is not None:
                self.residuals_type = res_type_prev

    def optimize_min(
        self,
        tol            = 3e-16,
        xtol           = None,
        ftol           = None,
        method         = 'Nelder-Mead',
        max_nfev       = None,
        residuals_type = None,
        rescale        = True,
        aid            = None,
        **kwargs
    ):
        res_type_prev = self.residuals_type
        aid = ensure_aid(aid)

        args = [
            'optimize_min_{arg}.MRF',
            'optimize_{arg}.MRF',
            'optimize_min_{arg}',
            'optimize_{arg}',
        ]
        xtol     = aid.hint_arg(xtol,     args, self.opt_xtol_default, arg = 'xtol')
        ftol     = aid.hint_arg(ftol,     args, self.opt_ftol_default, arg = 'ftol')
        max_nfev = aid.hint_arg(max_nfev, args, self.opt_nfev_default, arg = 'max_nfev')

        method  = method if method is not None else self.opt_method_default
        rescale = rescale if rescale is not None else self.opt_rescale_default

        if rescale:
            p_rescale = np.abs(np.asarray(self.parameters)) + aid.hint('MRF_rescale_min', 1e-5)
        else:
            p_rescale = np.asarray(1)

        if residuals_type is not None:
            self.residuals_type = residuals_type
        def opt(p):
            self.parameters = p * p_rescale
            return norm_sq(self.residuals_preferred)

        def opt_j(p):
            self.parameters = p
            jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
            return 2 * (
                jac.real.dot(self.residuals_preferred.real)
                + jac.imag.dot(self.residuals_preferred.imag)
            )

        def opt_h(p):
            self.parameters = p
            jac = self.residuals_jacobian * p_rescale.reshape(-1, 1)
            hess = 2 * (
                jac.real.dot(jac.T.real)
                + jac.imag.dot(jac.T.imag)
            )
            return hess

        more_args = dict()
        if method in ['Nelder-Mead']:
            #doesn't need jac or hess
            pass
        else:
            more_args['jac'] = opt_j
            more_args['hess'] = opt_h

        kwargs['maxiter'] = max_nfev
        #opt_r = least_squares(
        #TODO, this is here because matlab can't seem to import it.
        import scipy.optimize
        opt_r = scipy.optimize.minimize(
            opt,
            self.parameters,
            tol     = tol,
            method  = method,
            options = kwargs,
            **more_args
        )
        results = declarative.Bunch()
        results.update(opt_r)
        self.parameters = results.x * p_rescale
        aid.log(results.message, results.nfev, results.njev)
        if residuals_type is not None:
            self.residuals_type = res_type_prev
        return results
        return opt_r

    def eval_xfer(self, F_Hz):
        alt_filter = MultiReprFilter(
            parent = self,
            data = [],
            W    = [],
            F_Hz = F_Hz,
        )

        #TODO, this is a HACK, but needed for the parameterization
        alt_filter.F_data_max_Hz = self.F_data_max_Hz
        return alt_filter.xfer_fit

    def optimize_delay(
        self,
        only_delay = True,
        **kwargs
    ):
        #store the previous so they may be restored
        prev_hide_delay = self.gain_delay_coding.hide_delay
        prev_codings_ignore = self.codings_ignore

        #setup the delay to be fit
        self.gain_delay_coding.setup(hide_delay = False)
        #must clear other settings using the revision number
        self.codings_revision = self.codings_revision + 1

        if only_delay:
            #ignore all other codings
            self.codings_ignore = set(self.num_codings + self.den_codings)

        ret = self.optimize(**kwargs)

        #restore
        self.codings_ignore = prev_codings_ignore
        self.gain_delay_coding.setup(hide_delay = prev_hide_delay)
        return ret

    def codings_in_region(
        self,
        F_min_Hz,
        F_max_Hz,
        num = True,
        den = True,
    ):
        #TODO: add more tests and bandwidth limits
        gotten = []
        def test_get(coding):
            r_lst = coding.roots_c(self)
            if len(r_lst) == 1:
                r = r_lst[0]
                F = np.angle(r) / np.pi * self.F_nyquist_Hz
                if F_max_Hz is not None and F > F_max_Hz:
                    return
                if F_min_Hz is not None and F < F_min_Hz:
                    return
                gotten.append(coding)

        if num:
            for coding in self.num_codings:
                test_get(coding)
        if den:
            for coding in self.den_codings:
                test_get(coding)
        return gotten

    @contextlib.contextmanager
    def with_codings_only(self, codings, prev_ignores = True):
        prev_ignores = self.codings_ignore
        if prev_ignores:
            self.codings_ignore = self.codings_ignore.union(self.root_codings_set - set(codings))
        else:
            self.codings_ignore = self.root_codings_set - set(codings)
        yield
        self.codings_ignore = prev_ignores
        return

    @contextlib.contextmanager
    def without_codings_only(self, codings, prev_ignores = True):
        prev_ignores = self.codings_ignore
        if prev_ignores:
            self.codings_ignore = self.codings_ignore.union(set(codings))
        else:
            self.codings_ignore = set(codings)
        yield
        self.codings_ignore = prev_ignores


def check_jac(fitterMRF, idx):
    p_orig = np.asarray(fitterMRF.parameters)
    jac = fitterMRF.residuals_jacobian
    res1 = fitterMRF.residuals_preferred
    shift = [0] * len(p_orig)
    offset = .01 / (np.dot(jac[idx], jac[idx].conjugate()).real)**.5
    shift[idx] = offset
    fitterMRF.parameters = p_orig + np.asarray(shift)
    res2 = fitterMRF.residuals_preferred
    njac = (res2 - res1) / offset

    fitterMRF.parameters = p_orig
    return njac / jac[idx]

