# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np
import scipy
import declarative
import scipy.linalg

from .data_filtfit_base import DataFiltFitBase
from .svd import SVD_SV


def abs_sq(x):
    return x.real**2 + x.imag**2


class RationalDiscFilter(DataFiltFitBase):
    def __build__(
        self,
        _args = None,
        npoles      = None,
        nzeros      = None,
        ZPK         = ((), (), 1),
        mindelay    = False,
        stabilize   = False,
        parent      = None,
        extra_poles = None,
        extra_zeros = None,
        **kwargs
    ):
        if _args:
            raise RuntimeError("Only keyword arguments allowed")
        self.npoles    = npoles
        self.nzeros    = nzeros
        self.mindelay  = mindelay
        self.stabilize = stabilize

        if parent is not None:
            if extra_poles is None:
                extra_poles       = parent.extra_poles
            if extra_zeros is None:
                extra_zeros       = parent.extra_zeros
        else:
            if extra_poles is None:
                extra_poles = ()
            if extra_zeros is None:
                extra_zeros = ()

        self.extra_poles = extra_poles
        self.extra_zeros = extra_zeros

        super(RationalDiscFilter, self).__build__(
            parent = parent,
            **kwargs
        )

        #first to prevent lookups that shouldn't happen
        self.zeros     = ZPK[0]
        self.h_b_ref   = 1
        self.dependencies_for('zeros', 'b_vec')
        self.poles     = ZPK[1]
        self.h_a_ref   = 1
        self.dependencies_for('poles', 'a_vec')
        self.gain      = ZPK[2]

        @self.deco_generator
        def a_vec(self):
            """
            Setup so that the gain at the reference freq is ALWAYS 1
            """
            return self.poly.fromroots(self.poles).real

        @self.deco_setter
        def a_vec(self, val):
            gain = self.gain
            self.gain = gain / val[-1]
            self.h_a_ref = self.poly.val(self.Xp_ref, val) / val[-1]
            val = val / val[-1]
            if self.get_raw('stabilize'):
                poles = self.poly.roots(val)
                #use set raw so it won't update the gain redundantly/erroneously
                if np.any(abs(poles) > 1):
                    #only prevent using this vector if needed, stability is helped if we don't go convert between roots and vectors too much
                    #actually set it so that we get the stabilize action
                    self.poles = poles
                    return self.TAG_NO_SET
                else:
                    self.set_raw('poles', poles, setter = lambda s, x : x)

            self.dependencies('poles')
            return val

        @self.deco_generator
        def b_vec(self):
            """
            Setup so the gain is always specified at the ref
            """
            #TODO fix gain deps
            return self.get_raw('gain') * self.poly.fromroots(self.zeros).real

        @self.deco_setter
        def b_vec(self, val):
            val = np.asarray(val)
            gain = val[-1]
            self.gain = gain
            self.h_b_ref = self.poly.val(self.Xp_ref, val) / val[-1]
            if self.get_raw('mindelay'):
                zeros = self.poly.roots(val)
                if np.any(abs(zeros) > 1):
                    #only prevent using this vector if needed, stability is helped if we don't go convert between roots and vectors too much
                    #actually set it so that we get the stabilize action
                    self.zeros = zeros
                    return self.TAG_NO_SET
                else:
                    #use set raw so it won't update the gain redundantly/erroneously
                    self.set_raw('zeros', zeros, setter = lambda s, x : x)
            #since they had to be computed anyway for the gain adjustment
            self.dependencies('zeros')
            return val

        @self.deco_generator
        def zeros(self):
            return self.poly.roots(self.b_vec)

        @self.deco_setter
        def zeros(self, val):
            #get previous
            val = np.asarray(val)
            if self.get_raw('mindelay'):
                val = self.root_stabilize(
                    val,
                    method = 'flip',
                    real_neg = 'ignore',
                    real_pos = 'ignore',
                )
            self.dependencies('b_vec')
            h_ref_now = self.poly.valfromroots(self.Xp_ref, val)
            h_ref_prev = self.h_b_ref
            self.gain = self.get_raw('gain') * abs(h_ref_prev / h_ref_now)
            self.h_b_ref = h_ref_now
            return val

        @self.deco_generator
        def poles(self):
            return self.poly.roots(self.a_vec)

        @self.deco_setter
        def poles(self, val):
            val = np.asarray(val)
            if self.get_raw('stabilize'):
                val = self.root_stabilize(
                    val,
                    method = 'flip',
                    real_neg = 'ignore',
                    real_pos = 'ignore',
                )
            self.dependencies('a_vec')
            h_ref_now = self.poly.valfromroots(self.Xp_ref, val)
            h_ref_prev = self.h_a_ref
            self.gain = self.get_raw('gain') / abs(h_ref_prev / h_ref_now)
            self.h_a_ref = h_ref_now
            return val

        @self.deco_generator
        def zeros_full(self):
            return np.concatenate([self.zeros, self.extra_zeros])

        @self.deco_generator
        def poles_full(self):
            return np.concatenate([self.poles, self.extra_poles])

        @self.deco_setter
        def gain(self, val):
            #print("GAIN: ", val)
            return val

        @self.deco_generator
        def gain(self):
            raise RuntimeError("shouldn't compute")
            return self.b_vec[-1]/self.a_vec[-1]

        @self.deco_generator
        def V3(self):
            return self.poly.vander(self.Xp_grid, 3)

        @self.deco_generator
        def V_b(self):
            return self.poly.vander(self.Xp_grid, self.nzeros)

        @self.deco_generator
        def V_a(self):
            return self.poly.vander(self.Xp_grid, self.npoles)

        @self.deco_generator
        def h_delay(self):
            return np.exp(-2j * np.pi * self.F_Hz * self.delay_s)

        @self.deco_generator
        def h_a(self):
            #TODO account for poles_phase_adjust
            self.dependencies('poles', 'a_vec')
            X = self.Xp_grid
            pval = self.get_default('poles', None)
            #pval = self.poles
            if pval is not None:
                val = self.poly.valfromroots(X, pval)
                return val * self.Xn_grid**len(pval)

            a_vec = self.get_default('a_vec', None)
            if a_vec is not None:
                X = self.Xp_grid
                val = self.poly.val(X, a_vec)
                return val * self.Xn_grid**(len(a_vec) - 1)

            #otherwise compute it from the preferred poles representation
            pval = self.poles
            val = self.poly.valfromroots(X, pval)
            return val * self.Xn_grid**len(pval)

        @self.deco_generator
        def h_b(self):
            #TODO account for zeros_phase_adjust
            self.dependencies('zeros', 'b_vec', 'gain')
            X = self.Xp_grid
            zval = self.get_default('zeros', None)
            #zval = self.zeros
            if zval is not None:
                val = self.poly.valfromroots(X, zval)
                return self.gain * val * self.Xn_grid**len(zval)

            b_vec = self.get_default('b_vec', None)
            if b_vec is not None:
                #don't compute and normalize by h_ref since the b-coeffs carry the gain
                val = self.poly.val(X, b_vec)
                return val * self.Xn_grid**(len(b_vec) - 1)

            #otherwise compute it from the preferred poles representation
            zval = self.zeros
            val = self.poly.valfromroots(X, zval)
            return self.gain * val * self.Xn_grid**len(zval)

        @self.deco_generator
        def _extra_xfer_fit(self):
            if not self.extra_zeros and not self.extra_poles:
                return 1
            X = self.Xp_grid
            return self.poly.valfromroots(X, self.extra_zeros) / self.poly.valfromroots(X, self.extra_poles)

        @self.deco_generator
        def xfer_fit(self):
            return self._extra_xfer_fit * (self.h_b * self.h_delay) / self.h_a

        @self.deco_generator
        def residuals(self):
            debias_reweight = 1/(.001 + self.W**2)
            retB = declarative.Bunch()
            R = self.xfer_fit/self.data
            retB.resP = self.W * (R - 1)
            retB.resZ = self.W * (1/R - 1)
            retB.resD = self.W * (R - 1/R)
            retB.resD_total = np.sum(abs_sq(retB.resD)) / (2 * len(self.data))
            retB.total = np.sum(
                (abs(retB.resP)**2 + abs(retB.resZ * debias_reweight)**2) / (1 + debias_reweight)
            ) / (2 * len(self.data))
            return retB

        @self.deco_generator
        def residuals_total(self):
            return self.residuals.resD_total

        self.zeros_phase_adjust = None
        self.poles_phase_adjust = None

        @self.deco_generator
        def zeros_phasing(self):
            if self.zeros_phase_adjust is None:
                return self.Xn_grid**self.nzeros
            elif self.zeros_phase_adjust < 0:
                return self.Xn_grid**(self.nzeros + self.zeros_phase_adjust)
            else:
                return self.Xn_grid**(self.zeros_phase_adjust)
            return

        @self.deco_generator
        def poles_phasing(self):
            if self.poles_phase_adjust is None:
                return self.Xn_grid**self.npoles
            elif self.poles_phase_adjust < 0:
                return self.Xn_grid**(self.npoles + self.poles_phase_adjust)
            else:
                return self.Xn_grid**(self.poles_phase_adjust)
            return

        @self.deco_generator
        def A_z(self):
            return self.V_b * (
                self.W * self.zeros_phasing / (self.data / self.h_delay * self.h_a)
            ).reshape(-1, 1)

        @self.deco_generator
        def A_zp(self):
            return self.V_a * (
                self.W * self.zeros_phasing * self.h_b / (self.data / self.h_delay * self.h_a**2)
            ).reshape(-1, 1)

        @self.deco_generator
        def A_p(self):
            return self.V_a * (
                self.W * self.poles_phasing * (self.data / self.h_delay / self.h_b)
            ).reshape(-1, 1)

        @self.deco_generator
        def A_pz(self):
            return self.V_b * (
                self.W * self.poles_phasing * (self.data / self.h_delay * self.h_a / self.h_b**2)
            ).reshape(-1, 1)

        return  # ~__init__

    @property
    def ZPK(self):
        #so cheap to make that we don't try to autodelete
        #TODO, fix gain term as it is currently based on reference frequency
        return (self.zeros, self.poles, self.gain)

    def root_stabilize(
        self,
        roots,
        method = None,
        real_neg = None,
        real_pos = None,
    ):
        roots = np.asarray(roots)
        select = (abs(roots) > 1)
        if real_neg is None:
            pass
        elif real_neg == 'ignore':
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real < 0))
        elif real_neg == 'ignore_float':
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real < 0))
            float_sel = ((abs(roots.imag) < 1e-12) & (roots.real < -1) & (roots.real > -5))
            roots[float_sel] = -5
        else:
            raise RuntimeError("Bad Argument")

        if real_pos is None:
            pass
        elif real_pos == 'ignore':
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real > 0))
        elif real_pos == 'ignore_float':
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real > 0))
            float_sel = ((abs(roots.imag) < 1e-12) & (roots.real > 1) & (roots.real < 5))
            roots[float_sel] = 5
        else:
            raise RuntimeError("Bad Argument")

        if method is None:
            return roots
        elif method == 'flip':
            roots = np.copy(roots)
            roots[select] = 1 / roots[select].conjugate()
            return roots
        elif method == 'remove':
            #just nullify them
            roots = roots[~select]
            return roots
        elif method == 'zero':
            roots[select] = 0
        else:
            raise RuntimeError("Bad Argument")
        return roots

    def clear_unstable_poles(self):
        new_poles = []
        for r in self.poles:
            if abs(r) < 1:
                new_poles.append(r)
        self.poles = new_poles

    def clear_bigdelay_zeros(self):
        new_zeros = []
        for r in self.zeros:
            if abs(r) < 1:
                new_zeros.append(r)
        self.zeros = new_zeros

    def matched_pairs_clear(
        self,
        Q_rank_cutoff = .5,
    ):
        """
        Match unique closest pairs, if they are within a bandwidth of 0Hz, then they are ignored
        """
        poles_r, poles_c  = self.poles_split
        zeros_r, zeros_c = self.zeros_split

        def nearest_idx(lst_1, lst_2):
            nearest_lst = []
            for r1 in lst_1:
                if r1 is None:
                    nearest_lst.append(None)
                    continue
                dist_nearest = float('inf')
                idx_nearest = None
                for idx_2, r2 in enumerate(lst_2):
                    if r2 is None:
                        continue
                    dist = abs(r1 - r2)
                    if dist < dist_nearest:
                        idx_nearest = idx_2
                        dist_nearest = dist
                nearest_lst.append(idx_nearest)
            return nearest_lst
        z_nearest = nearest_idx(zeros_c, poles_c)
        p_nearest = nearest_idx(poles_c, zeros_c)

        z_duals = []
        p_duals = []
        for idx_z, idx_p in enumerate(z_nearest):
            if idx_p is None:
                continue
            if idx_z != p_nearest[idx_p]:
                #not a unique pairing
                continue
            z = zeros_c[idx_z]
            p = poles_c[idx_p]
            Q_rank = abs(p-z) * (1/(1 - abs(p))**2 + 1/(1 - abs(z))**2)**.5
            if Q_rank < Q_rank_cutoff:
                z_duals.append(idx_z)
                p_duals.append(idx_p)
        p_duals = set(p_duals)
        z_duals = set(z_duals)

        poles_new = []
        for idx_p, pole in enumerate(poles_c):
            if idx_p in p_duals:
                continue
            poles_new.append(pole)
            poles_new.append(pole.conjugate())
        zeros_new = []
        for idx_z, zero in enumerate(zeros_c):
            if idx_z in z_duals:
                continue
            zeros_new.append(zero)
            zeros_new.append(zero.conjugate())
        poles_new.extend(poles_r)
        zeros_new.extend(zeros_r)
        self.poles = poles_new
        self.zeros = zeros_new
        return

    def fit_poles_mod_zeros(
        self,
        max_size  = None,
        **kwargs
    ):
        #remove the effect of linear (delay) phasing
        A_z = self.A_z
        #A_z = np.hstack([(1j*F_Hz).reshape(-1, 1), A_z])
        q, r = np.linalg.qr(A_z)

        A_p = self.A_p
        #the str is because py2 doesn't like some symbol in there
        A_p = A_p - np.einsum(str('ij,jk->ik'), q, np.einsum(str('ij,ik->jk'), q.conjugate(), A_p))

        S, V = SVD_SV(
            np.vstack([A_p.real, A_p.imag]),
            n_smallest  = 3,
            overwrite_a = True,
        )
        #print("POLES SVD: ", S[-4:])
        self.a_vec = V.T[:, -1]
        return

    def fit_zeros_mod_poles(
        self,
        max_size  = None,
        **kwargs
    ):
        A_p = self.A_p
        #remove the effect of linear (delay) phasing
        #A_p = np.hstack([(1j*F_Hz).reshape(-1, 1), A_p])

        q, r = np.linalg.qr(A_p)

        A_z = self.A_z
        A_z = A_z - np.einsum(str('ij,jk->ik'), q, np.einsum(str('ij,ik->jk'), q.conjugate(), A_z))
        S, V = SVD_SV(
            np.vstack([A_z.real, A_z.imag]),
            n_smallest  = 3,
            overwrite_a = True,
        )
        #print("ZEROS SVD: ", S[-4:])
        self.b_vec = V.T[:, -1]
        return

    def match_pair_iter(
        self,
        Q_rank_cutoff,
        num_sequence = 2,
        zeros_first = True,
    ):
        npoles = self.npoles
        nzeros = self.nzeros

        self.matched_pairs_clear(Q_rank_cutoff = Q_rank_cutoff)

        self.npoles = len(self.poles)
        self.nzeros = len(self.zeros)

        if not zeros_first:
            self.fit_poles()
            num_sequence -= 1
        for idx in range(num_sequence):
            self.fit_zeros()
            self.fit_poles()

        if not zeros_first:
            self.fit_zeros()

        self.npoles = npoles
        self.nzeros = nzeros

    def fit_covar_x(self, sign = 1, p = True, z = True, **kwargs):
        #this version is indeed perturbative, so we have to modify differentially
        A_p = self.A_p
        A_z = self.A_z
        if sign > 0:
            new_res = (1 - self.data / self.xfer_fit) * self.W
        else:
            new_res = (1 - self.xfer_fit / self.data) * self.W
        #solve the problem with purely real taps
        a_fit, res, rank, s = scipy.linalg.lstsq(
            np.block([[A_p.real, -A_z.real], [A_p.imag, -A_z.imag]]),
            sign * np.hstack([new_res.real, new_res.imag]),
        )
        if p:
            self.a_vec = self.a_vec + a_fit[:A_p.shape[1]]
        if z:
            self.b_vec = self.b_vec + a_fit[A_p.shape[1]:]
        return

    def fit_covar_svd(self, p = True, z = True, **kwargs):
        #doesn't appear to be useful
        #this version is indeed perturbative, so we have to modify differentially
        A_p = self.A_p
        A_z = self.A_z
        S, V = SVD_SV(
            np.block([[A_p.real, -A_z.real], [A_p.imag, -A_z.imag]]),
            n_smallest  = 3,
            overwrite_a = True,
        )
        fit_vec = V.T[:, -1]
        if p:
            self.a_vec = self.a_vec + fit_vec[:A_p.shape[1]]
        if z:
            self.b_vec = self.b_vec + fit_vec[A_p.shape[1]:]
        return

    def clear_zp_above(self, bound = None):
        if bound is None:
            bound = max(self.F_Hz)
        p = self.poles
        F_p = (np.angle(p) / np.pi * self.F_nyquist_Hz + self.F_nyquist_Hz) % (2 * self.F_nyquist_Hz) - self.F_nyquist_Hz
        self.poles = self.poles[(abs(F_p) < bound) & (abs(p) > .0)]

        z = self.zeros
        F_z = (np.angle(z) / np.pi * self.F_nyquist_Hz + self.F_nyquist_Hz) % (2 * self.F_nyquist_Hz) - self.F_nyquist_Hz
        self.zeros = self.zeros[abs((F_z) < bound) & (abs(z) > .0)]
        return

    def fit_covar_p(self, sign = 1, p = True, z = False, **kwargs):
        #this version is indeed perturbative, so we have to modify differentially
        A_p = self.A_p
        A_z = self.A_pz * -1
        new_res = (1 - self.data / self.xfer_fit) * self.W
        #solve the problem with purely real taps
        a_fit, res, rank, s = scipy.linalg.lstsq(
            np.block([[A_p.real, -A_z.real], [A_p.imag, -A_z.imag]]),
            sign * np.hstack([new_res.real, new_res.imag]),
        )
        if p:
            self.a_vec = self.a_vec + a_fit[:A_p.shape[1]]
        if z:
            self.b_vec = self.b_vec + a_fit[A_p.shape[1]:]
        return

    def fit_covar_z(self, sign = 1, p = False, z = True, **kwargs):
        #this version is indeed perturbative, so we have to modify differentially
        A_z = self.A_z
        A_p = self.A_zp * 100
        new_res = (1 - self.xfer_fit / self.data) * self.W
        #solve the problem with purely real taps
        b_fit, res, rank, s = scipy.linalg.lstsq(
            np.block([[A_z.real, -A_p.real], [A_z.imag, -A_p.imag]]),
            sign * np.hstack([new_res.real, new_res.imag]),
        )
        if z:
            self.b_vec = self.b_vec + b_fit[:A_z.shape[1]]
        if p:
            self.a_vec = self.a_vec + b_fit[A_z.shape[1]:]
        return

    def fit_poles(self, **kwargs):
        #print(self.a_vec, self.b_vec)
        A_p = self.A_p
        #solve the problem with purely real taps
        a_fit, res, rank, s = scipy.linalg.lstsq(
            np.vstack([A_p.real, A_p.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        self.a_vec = a_fit
        return

    def fit_poles_d(self, **kwargs):
        A_p = self.A_p
        if len(self.a_vec) > self.npoles:
            self.a_vec = self.a_vec[:self.npoles]

        new_res = (1 - self.data / self.xfer_fit) * self.W
        a_fit, res, rank, s = scipy.linalg.lstsq(
            np.vstack([A_p.real, A_p.imag]),
            np.hstack([new_res.real, new_res.imag]),
        )
        prev_a = self.a_vec
        sdiff = (len(prev_a) - len(a_fit))
        if sdiff < 0:
            prev_a = np.append(prev_a, [0] * (-sdiff))
        elif sdiff > 0:
            a_fit = np.append(a_fit, [0] * (sdiff))
        self.a_vec = (prev_a + a_fit)
        return

    def fit_zeros(self, **kwargs):
        A_z = self.A_z
        b_fit, res, rank, s = scipy.linalg.lstsq(
            np.vstack([A_z.real, A_z.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        self.b_vec = b_fit
        return

    def fit_zeros_d(self, **kwargs):
        A_z = self.A_z
        if len(self.b_vec) > self.nzeros:
            self.b_vec = self.b_vec[:self.nzeros]
        new_res = (1 - self.xfer_fit / self.data) * self.W
        b_fit, res, rank, s = scipy.linalg.lstsq(
            np.vstack([A_z.real, A_z.imag]),
            np.hstack([new_res.real, new_res.imag]),
        )
        prev_b = self.b_vec
        sdiff = (len(prev_b) - len(b_fit))
        if sdiff < 0:
            prev_b = np.append(prev_b, [0] * (-sdiff))
        elif sdiff > 0:
            b_fit = np.append(b_fit, [0] * (sdiff))
        self.b_vec = (prev_b + b_fit)
        return

    def remove_doublets(self):
        p_r, p_c = self.poles_split
        z_r, z_c = self.zeros_split

        thresh_var = 1e-2
        thresh_covar = .1

        #mapping from zeros to poles, inside bunch with covar data
        used_zs = dict()

        for idx_p, p in enumerate(p_c):
            min_idx_z = None
            min_dist = 1
            for idx_z, z in enumerate(z_c):
                dist = abs_sq(p - z)
                if dist < min_dist:
                    min_idx_z = idx_z
                    min_dist = dist

            if min_idx_z is None:
                continue
            rp, rz, rcov_n = self.generate_ZPK_covar(
                cpoles_seq = [idx_p],
                czeros_seq = [min_idx_z]
            )
            if rp > thresh_var and rz > thresh_var and rcov_n > thresh_covar:
                if min_idx_z in used_zs:
                    if rcov_n < used_zs[min_idx_z].rcov_n:
                        #should maybe find second closest
                        continue

                print(rp, rz, rcov_n, min_dist**.5)
                used_zs[min_idx_z] = declarative.Bunch(
                    idx_p  = idx_p,
                    reff   = (rp**-2 + rz**-2)**(-.5),
                    rp     = rp,
                    rz     = rz,
                    rcov_n = rcov_n,
                )
        drop_ps = []
        drop_zs = []
        for idx_z, pB in used_zs.items():
            drop_zs.append(idx_z)
            drop_ps.append(pB.idx_p)

        drop_ps.sort()
        drop_zs.sort()

        def drop_join(roots, drop):
            c_r_join = []
            idx_r_prev = -1
            for idx_r in drop:
                c_r_join.append(roots[idx_r_prev+1:idx_r])
                idx_r_prev = idx_r
            c_r_join.append(roots[idx_r_prev+1:])
            from itertools import chain
            return tuple(chain(*c_r_join))

        p_c_new = drop_join(p_c, drop_ps)
        z_c_new = drop_join(z_c, drop_zs)
        self.poles = tuple(p_r) + p_c_new + tuple(r.conjugate() for r in p_c_new)
        self.zeros = tuple(z_r) + z_c_new + tuple(r.conjugate() for r in z_c_new)
        self.npoles = len(self.poles)
        self.nzeros = len(self.zeros)
        return

    def fit_pzpz(
        self,
        max_size          = None,
        collect_all       = False,
        zeros_first       = False,
        stabilize         = False,
        stabilize_svd     = True,
        mindelay          = False,
        mindelay_svd      = False,
        n_svd             = 1,
        n_iter            = 10,
    ):
        pre = declarative.Bunch()
        pre.mindelay  = self.mindelay
        pre.stabilize = self.stabilize

        collection = []
        if not zeros_first:
            if n_svd >= 1:
                fitA = self.fit_poles_mod_zeros
            else:
                fitA = self.fit_poles
            if n_svd >= 2:
                fitB = self.fit_zeros_mod_poles
            else:
                fitB = self.fit_zeros
            fitC = self.fit_poles
            fitD = self.fit_zeros
        else:
            if n_svd >= 1:
                fitA = self.fit_zeros_mod_poles
            else:
                fitA = self.fit_zeros
            if n_svd >= 2:
                fitB = self.fit_poles_mod_zeros
            else:
                fitB = self.fit_poles
            fitC = self.fit_zeros
            fitD = self.fit_poles

        self.mindelay  = mindelay_svd
        self.stabilize = stabilize_svd
        fitA(
            max_size  = max_size,
        )
        fitB(
            max_size = max_size,
        )
        self.mindelay  = mindelay
        self.stabilize = stabilize

        for i in range(n_iter - 1):
            fitC()
            fitD()
            if collect_all:
                collection.append(self.copy())

        if n_iter > 0:
            fitC()
            fitD()
        self.stabilize = pre.stabilize
        self.mindelay  = pre.mindelay
        if collect_all:
            collection.append(self.copy())
        return collection

    def fit_pz(
        self,
        n_iter = 0,
        n_svd = 1,
        **kwargs
    ):
        return self.fit_pzpz(
            n_iter = n_iter,
            n_svd  = n_svd,
            **kwargs
        )

    def eval_xfer(self, F_Hz):
        h_ref = self.poly.valfromroots(self.Xp_ref, self.zeros) / self.poly.valfromroots(self.Xp_ref, self.poles)
        #to reflect using Xp_grid
        X = np.exp(1j * np.pi * F_Hz / self.F_nyquist_Hz)
        h_val = self.poly.valfromroots(X, self.zeros) / self.poly.valfromroots(X, self.poles) * np.exp(-2j * np.pi * F_Hz * self.delay_s)
        return self.gain * h_val / abs(h_ref)


