# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np

from .base import (
    CodingType,
    Ipi,
    #I2pi
)
#import scipy.linalg


class CodingFA(CodingType):
    N_parameters   = 2
    p_F_Hz         = 0
    p_amplitude    = 0
    lock_F_Hz      = False
    lock_amplitude = False
    hide_F_Hz      = False
    hide_amplitude = False
    restabilize    = False

    def setup(
            self,
            hide_all       = None,
            lock_F_Hz      = None,
            hide_F_Hz      = None,
            lock_amplitude = None,
            hide_amplitude = None,
            disable        = None,
    ):
        if hide_all:
            hide_F_Hz = True
            hide_amplitude = True

        N_parameters = 2
        if lock_F_Hz is not None:
            self.lock_F_Hz = lock_F_Hz
        if hide_F_Hz is not None:
            self.hide_F_Hz = hide_F_Hz
        if self.hide_F_Hz:
            N_parameters -= 1

        if lock_amplitude is not None:
            self.lock_amplitude = lock_amplitude
        if hide_amplitude is not None:
            self.hide_amplitude = hide_amplitude
        if self.hide_amplitude:
            N_parameters -= 1

        if disable is not None:
            self.disable = disable
        if self.disable:
            N_parameters = 0

        self.N_parameters = N_parameters

    def update(self, A = None, B = None):
        if self.disable:
            assert(A is None and B is None)
            return
        if self.hide_amplitude:
            assert(B is None)
            if self.hide_F_Hz:
                assert(A is None)
            else:
                F_Hz = A
                if not self.lock_F_Hz:
                    self.p_F_Hz = F_Hz
        else:
            if self.hide_F_Hz:
                assert(B is None)
                amplitude = A
                if not self.lock_amplitude:
                    self.p_amplitude = amplitude
            else:
                F_Hz, amplitude = A, B
                if not self.lock_F_Hz:
                    self.p_F_Hz = F_Hz

                if not self.lock_amplitude:
                    self.p_amplitude = amplitude

    def reduce(self):
        if self.disable:
            return []
        if self.hide_amplitude:
            if self.hide_F_Hz:
                return []
            else:
                return [
                    self.p_F_Hz
                ]
        else:
            if self.hide_F_Hz:
                return [
                    self.p_amplitude
                ]
            else:
                return [
                    self.p_F_Hz,
                    self.p_amplitude,
                ]

    def update_roots(self, sys, r1):
        """
        """
        amp = abs(r1)
        if self.restabilize:
            if amp > 1:
                print("stabilizing: ", r1)
                amp = 1/amp
                self.lock_amplitude = True
        self.p_amplitude = amp
        self.p_F_Hz = np.angle(r1) / np.pi * sys.F_nyquist_Hz
        return

    @property
    def F_Hz(self):
        return self.p_F_Hz

    def transfer(self, sys):
        #frequency, linear amplitude
        amp = self.p_amplitude
        r   = amp * np.cos(np.pi * self.p_F_Hz / sys.F_nyquist_Hz)
        Xn   = sys.Xn_grid
        Xnsq = sys.Xn_grid_sq
        return (amp*amp) * Xnsq - 2 * Xn * r + 1

    def derivative(self, sys):
        if self.disable:
            return []
        #real/imaginary part of root
        amp = self.p_amplitude
        F_nyquist_Hz = sys.F_nyquist_Hz
        Xn   = sys.Xn_grid
        Xnsq = sys.Xn_grid_sq
        if not self.hide_F_Hz:
            if not self.hide_amplitude:
                rcos     = np.cos(np.pi * self.p_F_Hz / F_nyquist_Hz)
                rsin     = np.sin(np.pi * self.p_F_Hz / F_nyquist_Hz)
                return [
                    (2 * amp * np.pi / sys.F_nyquist_Hz) * (Xn * rsin) if not self.lock_F_Hz else 0,
                    ((2 * amp) * Xnsq - Xn * (2 * rcos)) if not self.lock_amplitude else 0,
                ]
            else:
                rsin     = np.sin(np.pi * self.p_F_Hz / F_nyquist_Hz)
                return [
                    (2 * amp * np.pi / sys.F_nyquist_Hz) * (Xn * rsin) if not self.lock_F_Hz else 0,
                ]
        else:
            if not self.hide_amplitude:
                rcos     = np.cos(np.pi * self.p_F_Hz / F_nyquist_Hz)
                return [
                    ((2 * amp) * Xnsq - Xn * (2 * rcos)) if not self.lock_amplitude else 0,
                ]
            else:
                return []

    def roots_c(self, sys):
        #frequency, linear amplitude
        return [self.p_amplitude * np.exp(Ipi * self.p_F_Hz / sys.F_nyquist_Hz)]


class CodingFAUX(CodingFA):
    restabilize = True
