# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np
from matplotlib import gridspec

from ..utilities.np import logspaced
from ..utilities.mpl import generate_stacked_plot_ax

from . import utilities


def plot_fit(
        self,
        fitter,
        plot_zp         = True,
        plot_to_nyquist = False,
        with_error      = True,
        with_data_error = True,
        plot_data       = True,
        N_points        = 1000,
        gs_base         = gridspec.GridSpec(1, 1)[0],
        fig             = None,
        ax_bunch        = None,
        xscale          = None,
):
    use_phase = True
    #if fitter.magsq_data:
    #    use_phase = False
    if ax_bunch is None:
        axB = generate_stacked_plot_ax(
            [
                ('mag', True),
                ('phase', use_phase),
            ],
            height_ratios = dict(
                mag = 2,
            ),
            heights_phys_in_default = 1,
            xscales = 'log',
            gs_base = gs_base,
            fig     = fig,
            ax_bunch = ax_bunch,
        )
    else:
        axB = ax_bunch
    if plot_zp:
        axB.mag2 = axB.mag.twinx()

    if plot_data:
        dmline = axB.mag.loglog(
            fitter.F_Hz,
            abs(fitter.data),
            label = 'Fnoise',
        )
        if axB.phase:
            dpline = axB.phase.semilogx(
                fitter.F_Hz,
                np.angle(fitter.data, deg = True),
                label = 'Fnoise',
            )
        if with_data_error:
            f = fitter.F_Hz
            h = fitter.data
            W = .001 + fitter.W
            select = (f > 0)
            min_err_lim = np.min(abs(h)) * max(1/3, np.min(1/(1 + 1 / W)))
            max_err_lim = np.max(abs(h)) * min(3, np.max((1 + 1 / W)))
            axB.mag.fill_between(
                fitter.F_Hz[select],
                np.maximum((abs(h) / (1 + 1 / W))[select], np.min(abs(h)) / 100),
                np.minimum((abs(h) * (1 + 1 / W))[select], np.max(abs(h)) * 100),
                color = dmline[0].get_color(),
                alpha = .2,
            )
            if axB.phase:
                axB.phase.fill_between(
                    f[select],
                    np.minimum((np.angle(h, deg = True) + 180/np.pi / W)[select], 180),
                    np.maximum((np.angle(h, deg = True) - 180/np.pi / W)[select], -180),
                    color = dpline[0].get_color(),
                    alpha = .2,
                )

    if plot_to_nyquist:
        f = np.sort(np.concatenate([
            np.linspace(0, fitter.F_nyquist_Hz, N_points//2),
            logspaced(np.min(fitter.F_Hz[fitter.F_Hz > 0]), fitter.F_nyquist_Hz, N_points//2),
        ]))
        h = fitter.eval_xfer(f)
    else:
        h = fitter.xfer_fit
        f = fitter.F_Hz

    mline = axB.mag.loglog(
        f,
        abs(h),
        label = 'fit {nP}P {nZ}Z Rsq:({res:.2e})'.format(
            nP = len(fitter.ZPK[1]),
            nZ = len(fitter.ZPK[0]),
            res     = fitter.residuals_total,
        ),
        #color = 'green',
    )
    if axB.phase:
        pline = axB.phase.semilogx(f, np.angle(h, deg = True), label = 'fit')
    #axB.resP.loglog(fitter.F_bal, abs(fitter.resP), label = 'fit')

    if with_error:
        select = (f > 0)
        axB.mag.fill_between(
            f[select],
            np.maximum((abs(h) / (1 + fitter.xfer_fit_error.mag_rel2))[select], np.min(abs(h)) / 10),
            np.minimum((abs(h) * (1 + fitter.xfer_fit_error.mag_rel2))[select], np.max(abs(h)) * 10),
            color = mline[0].get_color(),
            alpha = .3,
        )
        if axB.phase:
            axB.phase.fill_between(
                f[select],
                np.minimum((np.angle(h, deg = True) + 180/np.pi * fitter.xfer_fit_error.phase)[select], 180),
                np.maximum((np.angle(h, deg = True) - 180/np.pi * fitter.xfer_fit_error.phase)[select], -180),
                color = pline[0].get_color(),
                alpha = .3,
            )

    axB.mag.legend(
        ncol = 2,
        fontsize = 6,
        loc = 'lower left'
    )

    def plot_poles(ax):
        p = fitter.poles_full
        r = abs(p)
        theta = np.angle(p) / np.pi * fitter.F_nyquist_Hz
        select = abs(p) < 1
        ax.errorbar(
            theta[select],
            utilities.rlog(r[select], fitter.F_nyquist_Hz),
            xerr = (1-abs(r[select])) * fitter.F_nyquist_Hz,
            color = self.poleID_color,
            lw = .5,
            alpha = .5,
            marker = 'None',
            ls = 'None',
        )
        ax.scatter(
            theta[select],
            utilities.rlog(r[select], fitter.F_nyquist_Hz),
            s = 10,
            facecolor = self.poleID_color,
            linewidths = .5,
            #edgecolors = 'black',
        )
        ax.errorbar(
            theta[~select],
            utilities.rlog_neg(r[~select], fitter.F_nyquist_Hz),
            xerr = (1 - 1/abs(r[~select])) * fitter.F_nyquist_Hz,
            color = self.poleOD_color,
            lw = .5,
            alpha = .5,
            marker = 'None',
            ls = 'None',
        )
        ax.scatter(
            theta[~select],
            utilities.rlog_neg(r[~select], fitter.F_nyquist_Hz),
            s = 10,
            facecolor = self.poleOD_color,
            linewidths = .5,
            #edgecolors = 'black',
        )

    def plot_zeros(ax):
        z = fitter.zeros_full
        r = abs(z)
        theta = np.angle(z) / np.pi * fitter.F_nyquist_Hz
        select = abs(z) < 1

        ax.errorbar(
            theta[select],
            utilities.rlog(r[select], fitter.F_nyquist_Hz),
            xerr = (1-abs(r[select])) * fitter.F_nyquist_Hz,
            color = self.zeroID_color,
            lw = .5,
            alpha = .5,
            marker = 'None',
            ls = 'None',
        )
        ax.scatter(
            theta[select],
            utilities.rlog(r[select], fitter.F_nyquist_Hz),
            s = 10,
            facecolor = self.zeroID_color,
            linewidths = .5,
            #edgecolors = 'black',
            marker = 'x',
        )
        ax.errorbar(
            theta[~select],
            utilities.rlog_neg(r[~select], fitter.F_nyquist_Hz),
            xerr = (1 - 1/abs(r[~select])) * fitter.F_nyquist_Hz,
            color = self.zeroOD_color,
            lw = .5,
            alpha = .5,
            marker = 'None',
            ls = 'None',
        )
        ax.scatter(
            theta[~select],
            utilities.rlog_neg(r[~select], fitter.F_nyquist_Hz),
            s = 10,
            facecolor = self.zeroOD_color,
            linewidths = .5,
            #edgecolors = 'black',
            marker = 'x',
        )
    if plot_zp:
        plot_poles(axB.mag2)
        plot_zeros(axB.mag2)
        axB.mag2.set_ylabel('root bandwidth [Hz]')
        #axB.mag2.set_ylim(-2, None)
        axB.mag2.set_yscale('log_zoom')
        low, high = axB.mag2.get_ylim()
        axB.mag2.set_ylim(high, low)

    axB.mag.set_ylabel('Magnitude')
    if with_data_error:
        axB.mag.set_ylim(min_err_lim, max_err_lim)
    if axB.phase:
        axB.phase.set_ylabel('Phase [deg]')
    axB.ax_bottom.set_xlabel('Frequency [Hz]')

    axB.finalize()
    if xscale is None:
        midpt = (np.max(fitter.F_Hz) - np.min(fitter.F_Hz)) / 2
        if np.count_nonzero(fitter.F_Hz < midpt) < .65 * len(fitter.F_Hz):
            #then we conclude that the data is linearly spaced
            axB.ax_bottom.set_xscale('linear')
            axB.ax_bottom.set_xlim(min(fitter.F_Hz), max(fitter.F_Hz))
        else:
            axB.ax_bottom.set_xscale('log_zoom')
            select = fitter.F_Hz > 0
            axB.ax_bottom.set_xlim(min(fitter.F_Hz[select]), max(fitter.F_Hz))
    elif xscale == 'log' or xscale == 'log_zoom':
        select = fitter.F_Hz > 0
        axB.ax_bottom.set_xscale(xscale)
        axB.ax_bottom.set_xlim(min(fitter.F_Hz[select]), max(fitter.F_Hz))

    return axB
