"""
Utilities to manipulate ZPK roots S to/from Z, and make transfer functions
"""
from __future__ import division, print_function, unicode_literals
import numpy as np
import warnings

from iirrational.roots_bin import roots_bin_type


imag_tol = 1e-12
def TF_ZPK(F_Hz, ZPK, F_nyquist_Hz = None, group_len = 1):
    """
    Stably generates the ZPK transfer function, interlacing zeros and poles to prevent small number accumulating
    """
    Z, P, K = ZPK
    if F_nyquist_Hz is None:
        X = 1j * F_Hz
    else:
        X = np.exp(1j * np.pi * F_Hz / F_nyquist_Hz)

    if len(P) > len(Z):
        Z = list(Z) + [0] * (len(Z) - len(P))
    else:
        P = list(P) + [0] * (len(P) - len(Z))
    Z = np.asarray(Z)
    P = np.asarray(P)
    mlen = max(len(P), len(Z))
    h = 1
    for idx in range((mlen - 1) // group_len + 1):
        zs = Z[idx * group_len : (idx + 1) * group_len]
        ps = P[idx * group_len : (idx + 1) * group_len]
        h *= np.polynomial.polynomial.polyvalfromroots(X, zs)
        h /= np.polynomial.polynomial.polyvalfromroots(X, ps)
    return K * h


def ZtoS(ZPK, F_nyquist_Hz, F_ref_Hz):
    #use a list for mutability due to local/nonlocal referencing semantics
    gain_flip = [1]

    def roots_move(roots_list):
        moved = []
        for r in roots_list:
            if abs(r.imag) < imag_tol:
                if r < 0:
                    warnings.warn("Negative real root found in ZtoS, dropping")
                    gain_flip[0] *= -1
                else:
                    real_Hz = -(1 - r) * F_nyquist_Hz / np.pi
                    moved.append(real_Hz)
            else:
                F_Hz = np.angle(r) / np.pi * F_nyquist_Hz
                real_Hz = -(1 - abs(r)) * F_nyquist_Hz / np.pi
                moved.append(real_Hz + 1j * F_Hz)
        return moved
    Z_z, P_z, K_z = ZPK
    Z_s = roots_move(Z_z)
    P_s = roots_move(P_z)
    K_s = K_z * gain_flip[0] / abs(
        TF_ZPK(F_ref_Hz, (Z_s, P_s, 1), F_nyquist_Hz = None)
        / TF_ZPK(F_ref_Hz, (Z_z, P_z, 1), F_nyquist_Hz = F_nyquist_Hz)
    )
    return (Z_s, P_s, K_s)


def StoZ(ZPK, F_nyquist_Hz, F_ref_Hz):
    #use a list for mutability due to local/nonlocal referencing semantics
    gain_flip = [1]

    def roots_move(roots_list):
        moved = []
        for r in roots_list:
            if abs(r.imag / r.real) < imag_tol:
                real_D = 1 + r * np.pi / F_nyquist_Hz
                moved.append(real_D)
            else:
                if r.imag >= F_nyquist_Hz:
                    warnings.warn("Root Frequency above the nyquist encountered in StoZ, dropping")
                else:
                    F_D = np.exp(1j * np.pi * r.imag / F_nyquist_Hz)
                    amp_D = 1 + r.real * np.pi / F_nyquist_Hz
                    moved.append(amp_D * F_D)
        return moved
    Z_s, P_s, K_s = ZPK
    Z_z = roots_move(Z_s)
    P_z = roots_move(P_s)
    K_z = K_s * gain_flip[0] / abs(
        TF_ZPK(F_ref_Hz, (Z_z, P_z, 1), F_nyquist_Hz = F_nyquist_Hz)
        / TF_ZPK(F_ref_Hz, (Z_s, P_s, 1), F_nyquist_Hz = None)
    )
    return (Z_z, P_z, K_z)


def ZPK_fill(ZPK = None, Z = None, P = None, K = None):
    if Z is None:
        Z = ZPK[0]
    if P is None:
        P = ZPK[1]
    if K is None:
        K = ZPK[2]

    Z_r, Z_c, Z_u, pol = roots_bin_type(Z)
    assert(not Z_u)
    Z = np.concatenate([Z_r, Z_c, np.asarray(Z_c).conjugate()])
    P_r, P_c, P_u, pol = roots_bin_type(P)
    assert(not P_u)
    P = np.concatenate([P_r, P_c, np.asarray(P_c).conjugate()])

    return (Z, P, K)

