"""
Some utilities to help with dataset creation and annotation
"""
from __future__ import division, print_function, unicode_literals
import declarative

from ..annotate import padding_remove
from ..RDF import RationalDiscFilter

import numpy as np
from . import TF_utilities


def make_description(
    generator,
    sets        = 1,
    instances   = 1,
    description = None,
    annotation  = None,
    **kwargs
):
    doc = generator.__doc__
    if description is None:
        if doc is not None:
            description = padding_remove(doc)
        else:
            description = "<no description>"
    else:
        description = padding_remove(description)
        if annotation is None and doc is not None:
            annotation  = padding_remove(doc)
    return declarative.Bunch(
        generator   = generator,
        instances   = instances,
        description = description,
        annotation  = annotation,
        **kwargs
    )


def generator_autofill(
    F_Hz,
    SNR,
    F_nyquist_Hz,
    data               = None,
    bestfit_ZPK_z      = None,
    bestfit_ZPK_s      = None,
    residuals_log_best = None,
    F_ref_Hz           = 0,
    sN                 = 0,
    iN                 = 0,
    **kwargs
):
    if bestfit_ZPK_s:
        #replace with the ZPKs filled out fully
        bestfit_ZPK_s = TF_utilities.ZPK_fill(bestfit_ZPK_s)

    if bestfit_ZPK_z:
        #replace with the ZPKs filled out fully
        bestfit_ZPK_z = TF_utilities.ZPK_fill(bestfit_ZPK_z)

    if not bestfit_ZPK_z:
        if bestfit_ZPK_s:
            bestfit_ZPK_z = TF_utilities.StoZ(
                bestfit_ZPK_s,
                F_nyquist_Hz = F_nyquist_Hz,
                F_ref_Hz     = F_ref_Hz,
            )

    if not bestfit_ZPK_s:
        if bestfit_ZPK_z:
            bestfit_ZPK_s = TF_utilities.ZtoS(
                bestfit_ZPK_z,
                F_nyquist_Hz = F_nyquist_Hz,
                F_ref_Hz     = F_ref_Hz,
            )

    if data is None:
        if SNR is not None:
            rand = np.random.RandomState()
            rand.seed(iN)
            N = len(F_Hz)
            rel_noise = (rand.normal(1, 1/SNR, N) + 1j * rand.normal(0, 1/SNR, N))
        else:
            rel_noise = 1

        if bestfit_ZPK_s:
            data = rel_noise * TF_utilities.TF_ZPK(
                F_Hz,
                ZPK = bestfit_ZPK_s,
            )
        elif bestfit_ZPK_z:
            data = rel_noise * TF_utilities.TF_ZPK(
                F_Hz,
                ZPK = bestfit_ZPK_s,
                F_nyquist_Hz = F_nyquist_Hz,
            )

    if bestfit_ZPK_z is not None:
        fitter = RationalDiscFilter(
            data         = data,
            F_Hz         = F_Hz,
            W            = SNR,
            F_nyquist_Hz = F_nyquist_Hz,
            ZPK          = bestfit_ZPK_z,
        )
    else:
        fitter = None
    return declarative.Bunch(
        F_Hz               = F_Hz,
        data               = data,
        SNR                = SNR,
        F_nyquist_Hz       = F_nyquist_Hz,
        residuals_log_best = residuals_log_best,
        bestfit_ZPK_z      = bestfit_ZPK_z,
        bestfit_ZPK_s      = bestfit_ZPK_s,
        fitter             = fitter,
        F_ref_Hz           = F_ref_Hz,
        iN                 = iN,
        sN                 = sN,
        **kwargs
    )

