# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import numpy as np

from .. import MRF
from .. import codings
from ..fake_aid import ensure_aid


def flip_mindelay_opt(
        aid,
        doc_db = None,
):
    """
    Attempts to flip each non-mindelay zero and then reoptimize
    """
    aid = ensure_aid(aid)

    #TODO, also deal with real roots
    min_idx = 0

    while True:
        coding_lst = list(aid.fitter.num_codings)
        for idx, coding in enumerate(aid.fitter.num_codings):
            if idx <= min_idx:
                #TODO, change logic to not need this and be faster
                continue
            zeros = coding.roots_c(aid.fitter)
            if zeros:
                z = zeros[0]
                if abs(z) > 1:
                    aid.log("trying flipping", z)
                    coding_ins = codings.CodingFA()
                    #flip the root over, but also reduce its effect to help convergence
                    coding_ins.update_roots(aid.fitter, (1/z).conjugate())
                    coding_lst[idx] = coding_ins
                    min_idx = idx
                    break
        else:
            #breaks from the while loop only if for-loop break doesn't occur
            break

        fitter_new = MRF.MultiReprFilter(
            parent = aid.fitter,
            num_codings = coding_lst,
        )

        with fitter_new.with_codings_only([coding_ins]):
            fitter_new.optimize(aid = aid)
        fitter_new.optimize(aid = aid)

        aid.fitter_check(
            fitter_new,
            hint_name = 'mindelay_opt',
            variant   = 'OrdC',
        )
    return


def insert_triplets(
    aid,
    doc_db  = None,
    use_sos = False
):
    """
    iserts on each poles and zeros a complex and real roots with bandwidth 2x the data
    """
    aid = ensure_aid(aid)

    if not use_sos:
        cplx_t = codings.CodingnlFnlA
        real_t = codings.CodingRealnlA
    else:
        cplx_t = codings.CodingSOS
        real_t = codings.CodingSOS

    F_l_Hz = aid.fitter.F_data_max_Hz
    r_3x = 1 - 2 * F_l_Hz / aid.fitter.F_nyquist_Hz
    r_2x = 1 - 2 * F_l_Hz / aid.fitter.F_nyquist_Hz
    theta = .90 * F_l_Hz / aid.fitter.F_nyquist_Hz * np.pi

    coding_num_p = cplx_t()
    coding_num_p.update_roots(aid.fitter, r_3x * np.exp(1j * theta))
    coding_num_p2 = real_t()
    coding_num_p2.update_roots(aid.fitter, r_2x)
    coding_den_p2 = real_t()
    coding_den_p2.update_roots(aid.fitter, r_2x)
    coding_den_p = cplx_t()
    coding_den_p.update_roots(aid.fitter, r_3x * np.exp(1j * theta))

    fitter_new = MRF.MultiReprFilter(
        parent = aid.fitter,
        num_codings = aid.fitter.num_codings + [coding_num_p2],
        den_codings = aid.fitter.den_codings + [coding_den_p2],
    )
    res_pre = fitter_new.residuals_total
    with fitter_new.with_codings_only([coding_num_p, coding_num_p2] + [coding_den_p, coding_den_p2]):
        fitter_new.optimize(aid = aid)
    res_mid = fitter_new.residuals_total
    fitter_new.optimize(aid = aid)
    res_post = fitter_new.residuals_total

    res_ratio = fitter_new.residuals_total / aid.fitter.residuals_total

    aid.log("TRIPLETS (rat = {0}, pre = {1}, mid = {2}, post = {3}".format(res_ratio, res_pre, res_mid, res_post))
    return aid.fitter_check(
        fitter_new,
        hint_name = 'insert_triplets2',
        variant   = 'OrdUp',
    )


def set_min_BW(aid):
    aid = ensure_aid(aid)

    def modify_codings(codings):
        for coding in codings:
            roots = coding.roots_c(aid.fitter)
            if roots:
                r, = roots
                root_F_Hz = aid.fitter.F_nyquist_Hz * np.angle(r) / np.pi
                F_idx = np.searchsorted(aid.fitter.F_Hz, root_F_Hz)
                F_idx_low = max(0, F_idx - 2)
                F_idx_high = min(len(aid.fitter.F_Hz) - 1, F_idx + 2)
                #get the local mean width of bins to restrict bandwidths
                F_mean_Hz = (aid.fitter.F_Hz[F_idx_high] - aid.fitter.F_Hz[F_idx_low]) / (F_idx_high - F_idx_low)
                coding.option_set(aid.fitter, minimum_BW_Hz = F_mean_Hz * .5)
    modify_codings(aid.fitter.den_codings)
    modify_codings(aid.fitter.num_codings)

    aid.fitter_update()
    return
