# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np
#import copy
#import scipy

from .. import MRF_generate
from .. import annotate
from .. import plots
from ..utilities import args

from . import order_reduce
from . import disc_sequence
from . import fit_aid
from . import algorithms


def data2filter(
        argB            = None,
        data            = None,
        F_Hz            = None,
        SNR             = None,
        F_nyquist_Hz    = None,
        ZPK             = None,
        SNR_initial     = None,
        order_initial   = None,
        verbosity_limit = None,
        SNR_cutoff      = None,
        doc_db          = None,
        delay_s         = None,
        hints           = None,
):
    data            = args.argscan(locals(), argB, args.REQ, arg = 'data')
    F_Hz            = args.argscan(locals(), argB, args.REQ, arg = 'F_Hz')
    SNR             = args.argscan(locals(), argB, 1,        arg = 'SNR')
    F_nyquist_Hz    = args.argscan(locals(), argB, args.REQ, arg = 'F_nyquist_Hz')
    verbosity_limit = args.argscan(locals(), argB, 5,        arg = 'verbosity_limit')
    SNR_cutoff      = args.argscan(locals(), argB, 100,      arg = 'SNR_cutoff')
    SNR_initial     = args.argscan(locals(), argB, None,     arg = 'SNR_initial')
    ZPK             = args.argscan(locals(), argB, None,     arg = 'ZPK')
    order_initial   = args.argscan(locals(), argB, None,     arg = 'order_initial')
    delay_s         = args.argscan(locals(), argB, 0,        arg = 'delay_s')

    #TODO make annotate_setup
    doc_db = annotate.create(
        doc_db,
        name = "fit_sequence version 1",
        verbosity_limit = verbosity_limit
    )

    aid = fit_aid.FitAid(
        doc_db = doc_db,
        hints  = hints,
    )

    aid.hint_set('resavg_RthreshOrdDn', 1.10, default = True)
    aid.hint_set('resavg_RthreshOrdUp', 1.00, default = True)
    aid.hint_set('resavg_RthreshOrdC',  1.00, default = True)

    annotate.annotate(
        doc_db,
        name   = None,
        method = 'v1.fit_sequence',
        about = (
            """
            Version 1 smart fitter in iirrational library. Uses SVD method with high order over-fitting, then switches to nonlinear fits with heuristics to remove poles and zeros down to a reasonable system order.
            """
        ),
        verbosity = 3,
    )

    if SNR_initial is None:
        if SNR is None:
            SNR_initial = 1
        else:
            SNR_initial = np.minimum(SNR, SNR_cutoff)

    if ZPK is None:
        #initially ignore the weighting
        rat_fitter = disc_sequence.fit_rational_disc(
            F_Hz             = F_Hz,
            data             = data,
            SNR              = SNR_initial,
            nyquist_final_Hz = F_nyquist_Hz,
            order            = order_initial,
            aid              = aid,
            doc_db = annotate.annotate_into(
                doc_db,
                name = 'initial',
            ),
        )

        #TODO, add documentation
        aid.fitter_update(
            MRF_generate.MRF_generate(
                rat_fitter,
                delay_hide = True,
                coding_mapping = MRF_generate.coding_mapping_nl_safe,
            )
        )
    else:
        #TODO, add documentation
        aid.fitter_update(
            MRF_generate.MRF_generate_ZPK(
                rat_fitter,
                F_Hz             = F_Hz,
                data             = data,
                SNR              = SNR_initial,
                nyquist_final_Hz = F_nyquist_Hz,
                delay_hide       = True,
                coding_mapping   = MRF_generate.coding_mapping_nl_safe,
            )
        )

    aid.log("Initial Order: (Z= {0}, P= {1}, Z-P= {2})".format(
        len(aid.fitter.zeros),
        len(aid.fitter.poles),
        len(aid.fitter.zeros) - len(aid.fitter.poles),
    ))

    duals = order_reduce.match_pairs(aid = aid)
    #do a pre-reduce for speed with very conservative cutoff
    ORb = order_reduce.order_reduce(
        aid = aid,
        duals = duals,
        Q_rank_cutoff = .6,
    )
    aid.fitter_update(ORb.fitter_out)

    opt_r = aid.fitter.optimize(
        residuals_type = 'dualA',
        aid            = aid,
    )
    aid.fitter_update()

    opt_r = aid.fitter.optimize(aid = aid)
    aid.fitter_update()

    doc_sub = annotate.annotate(
        doc_db,
        name = 'nonlinear pre-reduce and optimize',
        fitter   = aid.fitter,
        plotter = plots.plot_fitter_flag,
        method = 'MultiReprFilter.optimize',
        about = (
            """
            initial conservative order reduction (for speed), followed by a nonlinear optimization.
            """
        ),
        verbosity = 3,
    )

    improved = algorithms.insert_triplets(
        aid    = aid,
        doc_db = doc_db,
    )
    if improved:
        improved = algorithms.insert_triplets(
            aid    = aid,
            doc_db = None,
        )

    opt_r = aid.fitter.optimize(aid = aid)
    aid.fitter_update()

    algorithms.set_min_BW(aid = aid)

    opt_r = aid.fitter.optimize(aid = aid)
    aid.fitter_update()

    doc_sub = annotate.annotate(
        doc_db,
        name = 'optimize nonlinear after bandwidth limiting',
        fitter   = aid.fitter,
        plotter = plots.plot_fitter_flag,
        method = 'MultiReprFilter.optimize',
        about = (
            """
            Root bandwidths limited in the nonlinear representation to half of the local average the frequency spacing.
            Nonlinear optimization then applied.
            """
        ),
        verbosity = 7,
    )

    order_reduce.order_reduce_aggressive(
        aid = aid,
        doc_db = annotate.annotate_into(
            doc_db,
            name = 'order_reduce 1',
        ),
    )

    order_reduce.order_reduce_negroots(
        aid = aid,
        doc_db = annotate.annotate_into(
            doc_db,
            name = 'remove_negroots',
        ),
    )
    aid.fitter_update()

    improved = True
    while improved:
        improved = order_reduce.order_reduce_weakroots(
            aid = aid,
            delay_s   = delay_s,
            doc_db = annotate.annotate_into(
                doc_db,
                name = 'remove_weakroots',
            ),
        )
        aid.fitter_update()
    order_reduce.order_reduce_negroots(
        aid      = aid,
        doc_db   = annotate.annotate_into(
            doc_db,
            name = 'remove_negroots',
        ),
    )

    opt_r = aid.fitter.optimize(
        aid            = aid,
        residuals_type = 'dualA',
    )
    aid.fitter_update()

    opt_r = aid.fitter.optimize(
        aid  = aid,
    )
    aid.fitter_update()

    order_reduce.order_reduce_negroots(
        aid      = aid,
        doc_db   = annotate.annotate_into(
            doc_db,
            name = 'remove_negroots',
        ),
    )

    if delay_s is None:
        aid.fitter.optimize_delay(aid = aid)
        aid.fitter_update()

    algorithms.flip_mindelay_opt(
        aid = aid,
        doc_db = annotate.annotate_into(
            doc_db,
            name = 'flip mindelay',
        ),
    )

    aid.fitter.optimize(aid = aid)
    aid.fitter_update()

    if delay_s is None:
        aid.fitter.optimize_delay(aid = aid)
        aid.fitter_update()
    aid.fitter.optimize(aid = aid)
    aid.fitter_update()

    if delay_s is None:
        aid.log("DELAY: ", aid.fitter.gain_delay_coding.p_delay)

    aid.log("FINAL RESIDUALS", aid.fitter.residuals_total)

    doc_sub = annotate.annotate(
        doc_db,
        name = 'final',
        fitter   = aid.fitter,
        plotter = plots.plot_fitter_flag,
        method = 'MultiReprFilter.optimize',
        about = (
            """
            Optimizes using nonlinear parameterizations,
            TODO: describe as used
            """
        ),
        verbosity = 0,
    )
    return aid


