# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
#import warnings

from .. import MRF
from .. import annotate


def fitter_order(fitter):
    return max(len(fitter.zeros), len(fitter.poles))


class FitterUnacceptable(Exception):
    pass


class FitAid(object):

    def __init__(
        self,
        doc_db,
        hints = None,
        residuals_type = 'log',
    ):
        self.fitter = None
        self.fitter_current    = None
        self.fitter_lowres_avg = None
        self.fitter_lowres_max = None
        self.fitter_lowres_med = None
        self.fitter_loword     = None

        self.residuals_type = residuals_type

        if hints is None:
            hints = dict()

        self.hints = dict(hints)
        self.hints_defaults = dict()

        self.doc_db = doc_db
        self.doc_stack = []

        self.section = [0]
        self.section_stack = []

        #TODO make this a hint
        self.verbosity = 5

    def _check_residuals(self, fitter):
        if not isinstance(fitter, MRF.MultiReprFilter):
            raise FitterUnacceptable()
        if fitter.residuals_type == self.residuals_type:
            if fitter.residuals_log_im_scale != 1:
                f_new = fitter.copy(
                    residuals_type = self.residuals_type,
                    residuals_log_im_scale = 1,
                )
                return f_new
        return None

    def hint_set(self, hname, hval, default = False):
        if default:
            self.hints.setdefault(hname, hval)
        else:
            self.hints[hname] = hval
        return

    def hint_arg(self, func_arg, *args, **kwargs):
        if func_arg is None:
            return self.hint(*args, **kwargs)
        else:
            return func_arg

    def hint(self, *args, **kwargs):
        superarg = []
        for arg in args[:-1]:
            if isinstance(arg, (list, tuple)):
                superarg.extend(arg)
            else:
                superarg.append(arg)
        if len(superarg) < 1:
            raise RuntimeError("Must have first hint and default return")

        default = args[-1]
        first_hint = superarg[0]

        lh_set = self.hints_defaults.get(first_hint, None)
        if lh_set is None:
            lh_set = set()
            self.hints_defaults[first_hint] = lh_set
        lh_set.add(default)

        for key in superarg:
            key = key.format(**kwargs)
            ret = self.hints.setdefault(key, None)
            if ret is not None:
                return ret
        return default

    def fitter_check(
            self,
            fitter_new,
            hint_name    = None,
            update       = True,
            variant = None,
    ):
        """
        """
        improved = False

        if variant is None:
            raise RuntimeError('Must specify variant for fitter_check, such as ["OrdUp", "OrdDn", "OrdC"]')

        if self.fitter is None:
            if update:
                self.fitter = fitter_new
                self.fitter_update()
            else:
                return True

        #TODO, should check that it is using standard residuals

        resavgN = lambda : fitter_new.residuals_total
        resmaxN = lambda : fitter_new.residuals_max
        resmedN = lambda : fitter_new.residuals_med
        resavgC = lambda : self.fitter_current.residuals_total
        resmaxC = lambda : self.fitter_current.residuals_max
        resmedC = lambda : self.fitter_current.residuals_med
        resavgB = lambda : self.fitter_lowres_avg.residuals_total
        resmaxB = lambda : self.fitter_lowres_max.residuals_max
        resmedB = lambda : self.fitter_lowres_med.residuals_med

        if hint_name is None:
            hint_name = 'default'

        resavg_Ethresh_hint = self.hint([
            'resavg_Ethresh{0}.{1}'.format(variant, hint_name),
            'resavg_Ethresh{0}'.format(variant),
        ], None)
        if resavg_Ethresh_hint is not None and resavgN() > resavg_Ethresh_hint:
            improved = False

        resmax_Ethresh_hint = self.hint([
            'resmax_Ethresh{0}.{1}'.format(variant, hint_name),
            'resmax_Ethresh{0}'.format(variant),
        ], None)
        if resmax_Ethresh_hint is not None and resmaxN() > resmax_Ethresh_hint:
            improved = False

        resavg_Rthresh_hint = self.hint([
            'resavg_Rthresh{0}.{1}'.format(variant, hint_name),
            'resavg_Rthresh{0}'.format(variant),
        ], None)
        if resavg_Rthresh_hint is not None and (resavgN() / resavgC()) > resavg_Rthresh_hint:
            improved = False

        resmax_Rthresh_hint = self.hint([
            'resmax_Rthresh{0}.{1}'.format(variant, hint_name),
            'resmax_Rthresh{0}'.format(variant),
        ], None)
        if resmax_Rthresh_hint is not None and (resmaxN() / resmaxC()) > resmax_Rthresh_hint:
            improved = False

        resmed_Rthresh_hint = self.hint([
            'resmed_Rthresh{0}.{1}'.format(variant, hint_name),
            'resmed_Rthresh{0}'.format(variant),
        ], None)
        if resmed_Rthresh_hint is not None and (resmedN() / resmedC()) > resmed_Rthresh_hint:
            improved = False

        resavgB_Rthresh_hint = self.hint([
            'resavgB_Rthresh{0}.{1}'.format(variant, hint_name),
            'resavgB_Rthresh{0}'.format(variant),
        ], None)
        if resavgB_Rthresh_hint is not None and (resavgN() / resavgB()) > resavgB_Rthresh_hint:
            improved = False

        resmaxB_Rthresh_hint = self.hint([
            'resmaxB_Rthresh{0}.{1}'.format(variant, hint_name),
            'resmaxB_Rthresh{0}'.format(variant),
        ], None)
        if resmaxB_Rthresh_hint is not None and (resmaxN() / resmaxB()) > resmaxB_Rthresh_hint:
            improved = False

        resmedB_Rthresh_hint = self.hint([
            'resmedB_Rthresh{0}.{1}'.format(variant, hint_name),
            'resmedB_Rthresh{0}'.format(variant),
        ], None)
        if resmedB_Rthresh_hint is not None and (resmedN() / resmedB()) > resmedB_Rthresh_hint:
            improved = False

        if improved and update:
            self.fitter_update(fitter_new)
        return improved

    def fitter_update(self, fitter = None):
        if fitter is None:
            fitter = self.fitter
        else:
            self.fitter = fitter

        try:
            fitter_use = self._check_residuals(fitter)
        except FitterUnacceptable:
            return
        if fitter_use is not None:
            fitter_use.optimize(aid = self)
            fitter_copy = [fitter_use]
        else:
            fitter_use = fitter
            fitter_copy = []

        def copy1():
            if not fitter_copy:
                fitter_copy.append(fitter.copy())
            return fitter_copy[0]

        new_res = fitter_use.residuals_total
        new_res_max = fitter_use.residuals_max
        new_res_med = fitter_use.residuals_med
        new_ord = fitter_order(fitter_use)

        if self.fitter_current is not None:
            old_res     = self.fitter_lowres_avg.residuals_total
            old_res_max = self.fitter_lowres_max.residuals_max
            old_res_med = self.fitter_lowres_med.residuals_med
            old_ord     = fitter_order(self.fitter_loword)

            if new_res < old_res:
                self.fitter_lowres_avg = copy1()

            if new_res_max < old_res_max:
                self.fitter_lowres_max = copy1()

            if new_res_med < old_res_med:
                self.fitter_lowres_med = copy1()

            if new_ord < old_ord:
                self.fitter_loword = copy1()

            self.fitter_current = copy1()
        else:
            self.fitter_current    = copy1()
            self.fitter_lowres_avg = copy1()
            self.fitter_lowres_max = copy1()
            self.fitter_lowres_med = copy1()
            self.fitter_loword     = copy1()
        return fitter_use

    def log(self, *args, **kwargs):
        #hint = kwargs.get('hint', None)

        if self.hint('no_report', False):
            return

        if self.hint('report_print', True):
            print(*args)

        #TODO put in the doc_db

    def digest_write(self, folder, format = 'markdown', **kwargs):
        if format in ['md', 'markdown']:
            return annotate.docdb2markdown(
                folder,
                self.doc_db,
                **kwargs
            )
        else:
            raise RuntimeError("Format Not Recognized, must be one of: ['md', 'markdown']")

