# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np
import declarative

from .. import MRF
from .. import codings
from ..fake_aid import ensure_aid


def insert_SOS(
    aid,
    doc_db = None,
):
    """
    inserts a dominant and non-dominant root. The dominant is a complex pair with the bandwidth 3x larger than the last data point
    the non-dominant is a bandwidth 3x larger, but a single real root
    """
    aid = ensure_aid(aid)

    F_l_Hz = aid.fitter.F_data_max_Hz
    r_3x = 1 - 2 * F_l_Hz / aid.fitter.F_nyquist_Hz
    r_2x = 1 - 2 * F_l_Hz / aid.fitter.F_nyquist_Hz
    theta = .8 * F_l_Hz / aid.fitter.F_nyquist_Hz * np.pi

    coding_num_p = codings.CodingnlFnlA()
    coding_num_p.update_roots(aid.fitter, r_3x * np.exp(1j * theta))
    coding_den_p = codings.CodingRealnlA()
    coding_den_p.update_roots(aid.fitter, r_2x)
    fitter_poles = MRF.MultiReprFilter(
        parent = aid.fitter,
        num_codings = aid.fitter.num_codings + [coding_num_p],
        den_codings = aid.fitter.den_codings + [coding_den_p],
    )
    fitter_poles.optimize(aid = aid)

    coding_den_z = codings.CodingnlFnlA()
    coding_den_z.update_roots(aid.fitter, r_3x * np.exp(1j * theta))
    coding_num_z = codings.CodingRealnlA()
    coding_num_z.update_roots(aid.fitter, r_2x)
    fitter_zeros = MRF.MultiReprFilter(
        parent = aid.fitter,
        num_codings = aid.fitter.num_codings + [coding_num_z],
        den_codings = aid.fitter.den_codings + [coding_den_z],
    )
    fitter_zeros.optimize(aid = aid)

    res_ratio_p = fitter_poles.residuals_total / aid.fitter.residuals_total
    res_ratio_z = fitter_zeros.residuals_total / aid.fitter.residuals_total
    aid.log("Triplet RATIO: ", res_ratio_p, res_ratio_z)
    #if res_ratio_p < 1 and res_ratio_z < 1:
    #    fitter_dual = MRF.MultiReprFilter(
    #        parent = aid.fitter,
    #        num_codings = aid.fitter.num_codings + [coding_num_p, coding_num_z],
    #        den_codings = aid.fitter.den_codings + [coding_den_p, coding_den_z],
    #    )
    #    fitter_dual.optimize(aid = aid)
    #    res_ratio_d = fitter_dual.residuals_total / aid.fitter.residuals_total
    #    if res_ratio_d < 1:
    #        return fitter_dual, True
    if res_ratio_z > 1 and res_ratio_p > 1:
        aid.log("Triplet Fit not improved", )
        aid.fitter_update()
        return False
    if res_ratio_p < res_ratio_z:
        aid.log("Triplet Fit IMPROVED by POLES", )
        aid.fitter_update(fitter_poles)
        return True
    else:
        aid.log("Triplet Fit IMPROVED by ZEROS", )
        aid.fitter_update(fitter_zeros)
        return True
    return aid.fitter


def DC_roots_sosify(
    aid,
    doc_db,
    desosify = False,
):
    """
    Find the roots which have a bandwidth larger than 4x the largest data point, and remove them
    """
    def SOSify(coding_list):
        new_list = []
        stored_reals    = []
        stored_cplx     = []
        codings_removed = []
        codings_added   = []
        for coding in coding_list:
            remove = False
            roots_r = np.asarray(coding.roots_r(fitterMRF))
            if len(roots_r):
                stored_reals.extend(roots_r)
                remove = True

            roots_c = np.asarray(coding.roots_c(fitterMRF))
            if len(roots_c):
                if remove:
                    #can't handle yet
                    assert(False)
                r = abs(roots_c)
                #get only the weak ones with bandwidth order the separation from zero
                select = (1 - r) > r.imag
                if np.any(select):
                    remove = True
                    #store the ones that aren't being changed
                    stored_cplx.extend(roots_c[~select])
                    for r in roots_c[select]:
                        new_coding = codings.CodingSOS()
                        new_coding.update_roots(fitterMRF, r)
                        codings_added.append(new_coding)
                        #print(roots_c, new_coding.roots(fitterMRF))
            if remove:
                codings_removed.append(coding)
            else:
                new_list.append(coding)

        while stored_cplx:
            new_coding = codings.CodingnlFnlA()
            new_coding.update_roots(fitterMRF, *stored_cplx[:1])
            stored_cplx[:1] = []
            codings_added.append(new_coding)

        while stored_reals:
            new_coding = codings.CodingSOS()
            new_coding.update_roots(fitterMRF, *stored_reals[:2])
            #print(stored_reals[:2], new_coding.roots(fitterMRF))
            stored_reals[:2] = []
            codings_added.append(new_coding)

        new_list = new_list + codings_added
        return declarative.Bunch(
            new_list        = new_list,
            codings_added   = codings_added,
            codings_removed = codings_removed,
        )

    def deSOSify(coding_list):
        new_list = []
        stored_reals    = []
        stored_cplx     = []
        codings_removed = []
        codings_added   = []
        bad_roots       = []
        for coding in coding_list:
            if not isinstance(coding, codings.CodingSOS):
                new_list.append(coding)
                continue
            #OK, now I know I have a SOS

            remove = False
            roots_r = np.asarray(coding.roots_r(fitterMRF))
            if len(roots_r) > 0:
                stored_reals.extend(roots_r)
                remove = True

            roots_c = np.asarray(coding.roots_c(fitterMRF))
            if len(roots_c) > 0:
                stored_cplx.extend(roots_c)
                remove = True

            if remove:
                codings_removed.append(coding)
            else:
                new_list.append(coding)

        for r in stored_cplx:
            if abs(r) > 1:
                bad_roots.append(r)
                bad_roots.append(r.conjugate())
            F_Hz = np.angle(r) / np.pi * fitterMRF.F_nyquist_Hz
            if F_Hz > fitterMRF.F_data_max_Hz:
                #skip it
                continue
            new_coding = codings.CodingnlFnlA()
            new_coding.update_roots(fitterMRF, r)
            codings_added.append(new_coding)

        for r in stored_reals:
            if r > 1 or r < -1:
                bad_roots.append(r)
                continue
            if r < 0:
                #just skip it
                continue
            new_coding = codings.CodingRealnlA()
            new_coding.update_roots(fitterMRF, r)
            codings_added.append(new_coding)

        new_list = new_list + codings_added
        return declarative.Bunch(
            new_list        = new_list,
            codings_added   = codings_added,
            codings_removed = codings_removed,
        )

    if not desosify:
        num_mod = SOSify(fitterMRF.num_codings)
        den_mod = SOSify(fitterMRF.den_codings)
    else:
        num_mod = deSOSify(fitterMRF.num_codings)
        den_mod = deSOSify(fitterMRF.den_codings)
    fitter_new = MRF.MultiReprFilter(
        parent = fitterMRF,
        num_codings = num_mod.new_list,
        den_codings = den_mod.new_list,
    )
    return fitter_new

